from mflux.models.common.lora.mapping.lora_mapping import LoRAMapping, LoRATarget
from mflux.models.common.lora.mapping.lora_transforms import LoraTransforms


class FluxLoRAMapping(LoRAMapping):
    @staticmethod
    def get_mapping() -> list[LoRATarget]:
        targets = []

        targets.extend(FluxLoRAMapping._get_standard_transformer_block_targets())
        targets.extend(FluxLoRAMapping._get_standard_single_transformer_block_targets())

        targets.extend(FluxLoRAMapping._get_bfl_transformer_block_targets())
        targets.extend(FluxLoRAMapping._get_bfl_single_transformer_block_targets())

        return targets

    @staticmethod
    def _get_standard_transformer_block_targets() -> list[LoRATarget]:
        return [
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_q",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_q.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.to_q.lora_B",
                    "transformer_blocks.{block}.attn.to_q.lora_up.weight",
                    "transformer_blocks.{block}.attn.to_q.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_q.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.to_q.lora_A",
                    "transformer_blocks.{block}.attn.to_q.lora_down.weight",
                    "transformer_blocks.{block}.attn.to_q.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_q.alpha",
                    "transformer_blocks.{block}.attn.to_q.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_k",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_k.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.to_k.lora_B",
                    "transformer_blocks.{block}.attn.to_k.lora_up.weight",
                    "transformer_blocks.{block}.attn.to_k.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_k.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.to_k.lora_A",
                    "transformer_blocks.{block}.attn.to_k.lora_down.weight",
                    "transformer_blocks.{block}.attn.to_k.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_k.alpha",
                    "transformer_blocks.{block}.attn.to_k.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_v",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_v.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.to_v.lora_B",
                    "transformer_blocks.{block}.attn.to_v.lora_up.weight",
                    "transformer_blocks.{block}.attn.to_v.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_v.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.to_v.lora_A",
                    "transformer_blocks.{block}.attn.to_v.lora_down.weight",
                    "transformer_blocks.{block}.attn.to_v.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_v.alpha",
                    "transformer_blocks.{block}.attn.to_v.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_out.0",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_out.0.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.to_out.0.lora_B",
                    "transformer_blocks.{block}.attn.to_out.0.lora_up.weight",
                    "transformer_blocks.{block}.attn.to_out.0.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_out.0.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.to_out.0.lora_A",
                    "transformer_blocks.{block}.attn.to_out.0.lora_down.weight",
                    "transformer_blocks.{block}.attn.to_out.0.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_out.0.alpha",
                    "transformer_blocks.{block}.attn.to_out.0.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.add_q_proj",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.attn.add_q_proj.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.add_q_proj.lora_B",
                    "transformer_blocks.{block}.attn.add_q_proj.lora_up.weight",
                    "transformer_blocks.{block}.attn.add_q_proj.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.attn.add_q_proj.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.add_q_proj.lora_A",
                    "transformer_blocks.{block}.attn.add_q_proj.lora_down.weight",
                    "transformer_blocks.{block}.attn.add_q_proj.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.attn.add_q_proj.alpha",
                    "transformer_blocks.{block}.attn.add_q_proj.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.add_k_proj",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.attn.add_k_proj.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.add_k_proj.lora_B",
                    "transformer_blocks.{block}.attn.add_k_proj.lora_up.weight",
                    "transformer_blocks.{block}.attn.add_k_proj.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.attn.add_k_proj.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.add_k_proj.lora_A",
                    "transformer_blocks.{block}.attn.add_k_proj.lora_down.weight",
                    "transformer_blocks.{block}.attn.add_k_proj.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.attn.add_k_proj.alpha",
                    "transformer_blocks.{block}.attn.add_k_proj.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.add_v_proj",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.attn.add_v_proj.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.add_v_proj.lora_B",
                    "transformer_blocks.{block}.attn.add_v_proj.lora_up.weight",
                    "transformer_blocks.{block}.attn.add_v_proj.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.attn.add_v_proj.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.add_v_proj.lora_A",
                    "transformer_blocks.{block}.attn.add_v_proj.lora_down.weight",
                    "transformer_blocks.{block}.attn.add_v_proj.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.attn.add_v_proj.alpha",
                    "transformer_blocks.{block}.attn.add_v_proj.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_add_out",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_add_out.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.to_add_out.lora_B",
                    "transformer_blocks.{block}.attn.to_add_out.lora_up.weight",
                    "transformer_blocks.{block}.attn.to_add_out.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_add_out.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.to_add_out.lora_A",
                    "transformer_blocks.{block}.attn.to_add_out.lora_down.weight",
                    "transformer_blocks.{block}.attn.to_add_out.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.attn.to_add_out.alpha",
                    "transformer_blocks.{block}.attn.to_add_out.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff.linear1",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.ff.net.0.proj.lora_B.weight",
                    "transformer.transformer_blocks.{block}.ff.net.0.proj.lora_B",
                    "transformer.transformer_blocks.{block}.ff.linear1.lora_B.weight",
                    "transformer.transformer_blocks.{block}.ff.linear1.lora_B",
                    "transformer_blocks.{block}.ff.net.0.proj.lora_up.weight",
                    "transformer_blocks.{block}.ff.linear1.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.ff.net.0.proj.lora_A.weight",
                    "transformer.transformer_blocks.{block}.ff.net.0.proj.lora_A",
                    "transformer.transformer_blocks.{block}.ff.linear1.lora_A.weight",
                    "transformer.transformer_blocks.{block}.ff.linear1.lora_A",
                    "transformer_blocks.{block}.ff.net.0.proj.lora_down.weight",
                    "transformer_blocks.{block}.ff.linear1.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.ff.net.0.proj.alpha",
                    "transformer.transformer_blocks.{block}.ff.linear1.alpha",
                    "transformer_blocks.{block}.ff.net.0.proj.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff.linear2",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.ff.net.2.lora_B.weight",
                    "transformer.transformer_blocks.{block}.ff.net.2.lora_B",
                    "transformer.transformer_blocks.{block}.ff.linear2.lora_B.weight",
                    "transformer.transformer_blocks.{block}.ff.linear2.lora_B",
                    "transformer_blocks.{block}.ff.net.2.lora_up.weight",
                    "transformer_blocks.{block}.ff.linear2.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.ff.net.2.lora_A.weight",
                    "transformer.transformer_blocks.{block}.ff.net.2.lora_A",
                    "transformer.transformer_blocks.{block}.ff.linear2.lora_A.weight",
                    "transformer.transformer_blocks.{block}.ff.linear2.lora_A",
                    "transformer_blocks.{block}.ff.net.2.lora_down.weight",
                    "transformer_blocks.{block}.ff.linear2.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.ff.net.2.alpha",
                    "transformer.transformer_blocks.{block}.ff.linear2.alpha",
                    "transformer_blocks.{block}.ff.net.2.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff_context.linear1",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.ff_context.net.0.proj.lora_B.weight",
                    "transformer.transformer_blocks.{block}.ff_context.net.0.proj.lora_B",
                    "transformer.transformer_blocks.{block}.ff_context.linear1.lora_B.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear1.lora_B",
                    "transformer_blocks.{block}.ff_context.net.0.proj.lora_up.weight",
                    "transformer_blocks.{block}.ff_context.linear1.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.ff_context.net.0.proj.lora_A.weight",
                    "transformer.transformer_blocks.{block}.ff_context.net.0.proj.lora_A",
                    "transformer.transformer_blocks.{block}.ff_context.linear1.lora_A.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear1.lora_A",
                    "transformer_blocks.{block}.ff_context.net.0.proj.lora_down.weight",
                    "transformer_blocks.{block}.ff_context.linear1.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.ff_context.net.0.proj.alpha",
                    "transformer.transformer_blocks.{block}.ff_context.linear1.alpha",
                    "transformer_blocks.{block}.ff_context.net.0.proj.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff_context.linear2",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.ff_context.net.2.lora_B.weight",
                    "transformer.transformer_blocks.{block}.ff_context.net.2.lora_B",
                    "transformer.transformer_blocks.{block}.ff_context.linear2.lora_B.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear2.lora_B",
                    "transformer_blocks.{block}.ff_context.net.2.lora_up.weight",
                    "transformer_blocks.{block}.ff_context.linear2.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.ff_context.net.2.lora_A.weight",
                    "transformer.transformer_blocks.{block}.ff_context.net.2.lora_A",
                    "transformer.transformer_blocks.{block}.ff_context.linear2.lora_A.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear2.lora_A",
                    "transformer_blocks.{block}.ff_context.net.2.lora_down.weight",
                    "transformer_blocks.{block}.ff_context.linear2.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.ff_context.net.2.alpha",
                    "transformer.transformer_blocks.{block}.ff_context.linear2.alpha",
                    "transformer_blocks.{block}.ff_context.net.2.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.norm1.linear",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.norm1.linear.lora_B.weight",
                    "transformer.transformer_blocks.{block}.norm1.linear.lora_B",
                    "transformer_blocks.{block}.norm1.linear.lora_up.weight",
                    "transformer_blocks.{block}.norm1.linear.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.norm1.linear.lora_A.weight",
                    "transformer.transformer_blocks.{block}.norm1.linear.lora_A",
                    "transformer_blocks.{block}.norm1.linear.lora_down.weight",
                    "transformer_blocks.{block}.norm1.linear.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.norm1.linear.alpha",
                    "transformer_blocks.{block}.norm1.linear.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.norm1_context.linear",
                possible_up_patterns=[
                    "transformer.transformer_blocks.{block}.norm1_context.linear.lora_B.weight",
                    "transformer.transformer_blocks.{block}.norm1_context.linear.lora_B",
                    "transformer_blocks.{block}.norm1_context.linear.lora_up.weight",
                    "transformer_blocks.{block}.norm1_context.linear.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.transformer_blocks.{block}.norm1_context.linear.lora_A.weight",
                    "transformer.transformer_blocks.{block}.norm1_context.linear.lora_A",
                    "transformer_blocks.{block}.norm1_context.linear.lora_down.weight",
                    "transformer_blocks.{block}.norm1_context.linear.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.transformer_blocks.{block}.norm1_context.linear.alpha",
                    "transformer_blocks.{block}.norm1_context.linear.alpha",
                ],
            ),
        ]

    @staticmethod
    def _get_standard_single_transformer_block_targets() -> list[LoRATarget]:
        return [
            LoRATarget(
                model_path="single_transformer_blocks.{block}.attn.to_q",
                possible_up_patterns=[
                    "transformer.single_transformer_blocks.{block}.attn.to_q.lora_B.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_q.lora_B",
                    "single_transformer_blocks.{block}.attn.to_q.lora_up.weight",
                    "single_transformer_blocks.{block}.attn.to_q.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.single_transformer_blocks.{block}.attn.to_q.lora_A.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_q.lora_A",
                    "single_transformer_blocks.{block}.attn.to_q.lora_down.weight",
                    "single_transformer_blocks.{block}.attn.to_q.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.single_transformer_blocks.{block}.attn.to_q.alpha",
                    "single_transformer_blocks.{block}.attn.to_q.alpha",
                ],
            ),
            LoRATarget(
                model_path="single_transformer_blocks.{block}.attn.to_k",
                possible_up_patterns=[
                    "transformer.single_transformer_blocks.{block}.attn.to_k.lora_B.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_k.lora_B",
                    "single_transformer_blocks.{block}.attn.to_k.lora_up.weight",
                    "single_transformer_blocks.{block}.attn.to_k.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.single_transformer_blocks.{block}.attn.to_k.lora_A.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_k.lora_A",
                    "single_transformer_blocks.{block}.attn.to_k.lora_down.weight",
                    "single_transformer_blocks.{block}.attn.to_k.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.single_transformer_blocks.{block}.attn.to_k.alpha",
                    "single_transformer_blocks.{block}.attn.to_k.alpha",
                ],
            ),
            LoRATarget(
                model_path="single_transformer_blocks.{block}.attn.to_v",
                possible_up_patterns=[
                    "transformer.single_transformer_blocks.{block}.attn.to_v.lora_B.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_v.lora_B",
                    "single_transformer_blocks.{block}.attn.to_v.lora_up.weight",
                    "single_transformer_blocks.{block}.attn.to_v.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.single_transformer_blocks.{block}.attn.to_v.lora_A.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_v.lora_A",
                    "single_transformer_blocks.{block}.attn.to_v.lora_down.weight",
                    "single_transformer_blocks.{block}.attn.to_v.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.single_transformer_blocks.{block}.attn.to_v.alpha",
                    "single_transformer_blocks.{block}.attn.to_v.alpha",
                ],
            ),
            LoRATarget(
                model_path="single_transformer_blocks.{block}.proj_mlp",
                possible_up_patterns=[
                    "transformer.single_transformer_blocks.{block}.proj_mlp.lora_B.weight",
                    "transformer.single_transformer_blocks.{block}.proj_mlp.lora_B",
                    "single_transformer_blocks.{block}.proj_mlp.lora_up.weight",
                    "single_transformer_blocks.{block}.proj_mlp.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.single_transformer_blocks.{block}.proj_mlp.lora_A.weight",
                    "transformer.single_transformer_blocks.{block}.proj_mlp.lora_A",
                    "single_transformer_blocks.{block}.proj_mlp.lora_down.weight",
                    "single_transformer_blocks.{block}.proj_mlp.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.single_transformer_blocks.{block}.proj_mlp.alpha",
                    "single_transformer_blocks.{block}.proj_mlp.alpha",
                ],
            ),
            LoRATarget(
                model_path="single_transformer_blocks.{block}.proj_out",
                possible_up_patterns=[
                    "transformer.single_transformer_blocks.{block}.proj_out.lora_B.weight",
                    "transformer.single_transformer_blocks.{block}.proj_out.lora_B",
                    "single_transformer_blocks.{block}.proj_out.lora_up.weight",
                    "single_transformer_blocks.{block}.proj_out.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.single_transformer_blocks.{block}.proj_out.lora_A.weight",
                    "transformer.single_transformer_blocks.{block}.proj_out.lora_A",
                    "single_transformer_blocks.{block}.proj_out.lora_down.weight",
                    "single_transformer_blocks.{block}.proj_out.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.single_transformer_blocks.{block}.proj_out.alpha",
                    "single_transformer_blocks.{block}.proj_out.alpha",
                ],
            ),
            LoRATarget(
                model_path="single_transformer_blocks.{block}.norm.linear",
                possible_up_patterns=[
                    "transformer.single_transformer_blocks.{block}.norm.linear.lora_B.weight",
                    "transformer.single_transformer_blocks.{block}.norm.linear.lora_B",
                    "single_transformer_blocks.{block}.norm.linear.lora_up.weight",
                    "single_transformer_blocks.{block}.norm.linear.lora_B.weight",
                ],
                possible_down_patterns=[
                    "transformer.single_transformer_blocks.{block}.norm.linear.lora_A.weight",
                    "transformer.single_transformer_blocks.{block}.norm.linear.lora_A",
                    "single_transformer_blocks.{block}.norm.linear.lora_down.weight",
                    "single_transformer_blocks.{block}.norm.linear.lora_A.weight",
                ],
                possible_alpha_patterns=[
                    "transformer.single_transformer_blocks.{block}.norm.linear.alpha",
                    "single_transformer_blocks.{block}.norm.linear.alpha",
                ],
            ),
        ]

    @staticmethod
    def _get_bfl_transformer_block_targets() -> list[LoRATarget]:
        return [
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_q",
                possible_up_patterns=["lora_unet_double_blocks_{block}_img_attn_qkv.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_img_attn_qkv.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_img_attn_qkv.alpha"],
                up_transform=LoraTransforms.split_q_up,
                down_transform=LoraTransforms.split_q_down,
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_k",
                possible_up_patterns=["lora_unet_double_blocks_{block}_img_attn_qkv.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_img_attn_qkv.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_img_attn_qkv.alpha"],
                up_transform=LoraTransforms.split_k_up,
                down_transform=LoraTransforms.split_k_down,
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_v",
                possible_up_patterns=["lora_unet_double_blocks_{block}_img_attn_qkv.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_img_attn_qkv.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_img_attn_qkv.alpha"],
                up_transform=LoraTransforms.split_v_up,
                down_transform=LoraTransforms.split_v_down,
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_out.0",
                possible_up_patterns=["lora_unet_double_blocks_{block}_img_attn_proj.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_img_attn_proj.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_img_attn_proj.alpha"],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.add_q_proj",
                possible_up_patterns=["lora_unet_double_blocks_{block}_txt_attn_qkv.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_txt_attn_qkv.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_txt_attn_qkv.alpha"],
                up_transform=LoraTransforms.split_q_up,
                down_transform=LoraTransforms.split_q_down,
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.add_k_proj",
                possible_up_patterns=["lora_unet_double_blocks_{block}_txt_attn_qkv.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_txt_attn_qkv.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_txt_attn_qkv.alpha"],
                up_transform=LoraTransforms.split_k_up,
                down_transform=LoraTransforms.split_k_down,
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.add_v_proj",
                possible_up_patterns=["lora_unet_double_blocks_{block}_txt_attn_qkv.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_txt_attn_qkv.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_txt_attn_qkv.alpha"],
                up_transform=LoraTransforms.split_v_up,
                down_transform=LoraTransforms.split_v_down,
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_add_out",
                possible_up_patterns=["lora_unet_double_blocks_{block}_txt_attn_proj.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_txt_attn_proj.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_txt_attn_proj.alpha"],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff.linear1",
                possible_up_patterns=["lora_unet_double_blocks_{block}_img_mlp_0.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_img_mlp_0.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_img_mlp_0.alpha"],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff.linear2",
                possible_up_patterns=["lora_unet_double_blocks_{block}_img_mlp_2.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_img_mlp_2.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_img_mlp_2.alpha"],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff_context.linear1",
                possible_up_patterns=["lora_unet_double_blocks_{block}_txt_mlp_0.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_txt_mlp_0.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_txt_mlp_0.alpha"],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff_context.linear2",
                possible_up_patterns=["lora_unet_double_blocks_{block}_txt_mlp_2.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_txt_mlp_2.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_txt_mlp_2.alpha"],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.norm1.linear",
                possible_up_patterns=["lora_unet_double_blocks_{block}_img_mod_lin.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_img_mod_lin.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_img_mod_lin.alpha"],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.norm1_context.linear",
                possible_up_patterns=["lora_unet_double_blocks_{block}_txt_mod_lin.lora_up.weight"],
                possible_down_patterns=["lora_unet_double_blocks_{block}_txt_mod_lin.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_double_blocks_{block}_txt_mod_lin.alpha"],
            ),
        ]

    @staticmethod
    def _get_bfl_single_transformer_block_targets() -> list[LoRATarget]:
        return [
            LoRATarget(
                model_path="single_transformer_blocks.{block}.attn.to_q",
                possible_up_patterns=["lora_unet_single_blocks_{block}_linear1.lora_up.weight"],
                possible_down_patterns=["lora_unet_single_blocks_{block}_linear1.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_single_blocks_{block}_linear1.alpha"],
                up_transform=LoraTransforms.split_single_q_up,
                down_transform=LoraTransforms.split_single_q_down,
            ),
            LoRATarget(
                model_path="single_transformer_blocks.{block}.attn.to_k",
                possible_up_patterns=["lora_unet_single_blocks_{block}_linear1.lora_up.weight"],
                possible_down_patterns=["lora_unet_single_blocks_{block}_linear1.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_single_blocks_{block}_linear1.alpha"],
                up_transform=LoraTransforms.split_single_k_up,
                down_transform=LoraTransforms.split_single_k_down,
            ),
            LoRATarget(
                model_path="single_transformer_blocks.{block}.attn.to_v",
                possible_up_patterns=["lora_unet_single_blocks_{block}_linear1.lora_up.weight"],
                possible_down_patterns=["lora_unet_single_blocks_{block}_linear1.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_single_blocks_{block}_linear1.alpha"],
                up_transform=LoraTransforms.split_single_v_up,
                down_transform=LoraTransforms.split_single_v_down,
            ),
            LoRATarget(
                model_path="single_transformer_blocks.{block}.proj_mlp",
                possible_up_patterns=["lora_unet_single_blocks_{block}_linear1.lora_up.weight"],
                possible_down_patterns=["lora_unet_single_blocks_{block}_linear1.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_single_blocks_{block}_linear1.alpha"],
                up_transform=LoraTransforms.split_single_mlp_up,
                down_transform=LoraTransforms.split_single_mlp_down,
            ),
            LoRATarget(
                model_path="single_transformer_blocks.{block}.proj_out",
                possible_up_patterns=["lora_unet_single_blocks_{block}_linear2.lora_up.weight"],
                possible_down_patterns=["lora_unet_single_blocks_{block}_linear2.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_single_blocks_{block}_linear2.alpha"],
            ),
            LoRATarget(
                model_path="single_transformer_blocks.{block}.norm.linear",
                possible_up_patterns=["lora_unet_single_blocks_{block}_modulation_lin.lora_up.weight"],
                possible_down_patterns=["lora_unet_single_blocks_{block}_modulation_lin.lora_down.weight"],
                possible_alpha_patterns=["lora_unet_single_blocks_{block}_modulation_lin.alpha"],
            ),
        ]



import mlx.core as mx
from mlx import nn


class QwenFeedForward(nn.Module):
    def __init__(self, dim: int = 3072):
        super().__init__()
        self.mlp_in = nn.Linear(dim, 4 * dim, bias=True)
        self.mlp_out = nn.Linear(4 * dim, dim, bias=True)

    def __call__(self, hidden_states: mx.array) -> mx.array:
        hidden_states = self.mlp_in(hidden_states)
        hidden_states = nn.gelu_approx(hidden_states)
        hidden_states = self.mlp_out(hidden_states)
        return hidden_states



from typing import List

from mflux.models.common.weights.mapping.weight_mapping import WeightMapping, WeightTarget
from mflux.models.common.weights.mapping.weight_transforms import WeightTransforms


class SeedVR2WeightMapping(WeightMapping):
    @staticmethod
    def get_transformer_mapping() -> List[WeightTarget]:
        num_blocks = 32
        return [
            WeightTarget(
                to_pattern="vid_in.proj.weight",
                from_pattern=["vid_in.proj.weight"],
            ),
            WeightTarget(
                to_pattern="vid_in.proj.bias",
                from_pattern=["vid_in.proj.bias"],
            ),
            WeightTarget(
                to_pattern="txt_in.weight",
                from_pattern=["txt_in.weight"],
            ),
            WeightTarget(
                to_pattern="txt_in.bias",
                from_pattern=["txt_in.bias"],
            ),
            WeightTarget(
                to_pattern="emb_in.proj_in.weight",
                from_pattern=["emb_in.proj_in.weight"],
            ),
            WeightTarget(
                to_pattern="emb_in.proj_in.bias",
                from_pattern=["emb_in.proj_in.bias"],
            ),
            WeightTarget(
                to_pattern="emb_in.proj_hid.weight",
                from_pattern=["emb_in.proj_hid.weight"],
            ),
            WeightTarget(
                to_pattern="emb_in.proj_hid.bias",
                from_pattern=["emb_in.proj_hid.bias"],
            ),
            WeightTarget(
                to_pattern="emb_in.proj_out.weight",
                from_pattern=["emb_in.proj_out.weight"],
            ),
            WeightTarget(
                to_pattern="emb_in.proj_out.bias",
                from_pattern=["emb_in.proj_out.bias"],
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.proj_qkv_vid.weight",
                from_pattern=[
                    "blocks.{block}.attn.proj_qkv.vid.weight",
                    "blocks.{block}.attn.proj_qkv.all.weight",
                ],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.proj_qkv_txt.weight",
                from_pattern=[
                    "blocks.{block}.attn.proj_qkv.txt.weight",
                    "blocks.{block}.attn.proj_qkv.all.weight",
                ],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.norm_q_vid.weight",
                from_pattern=[
                    "blocks.{block}.attn.norm_q.vid.weight",
                    "blocks.{block}.attn.norm_q.all.weight",
                ],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.norm_q_txt.weight",
                from_pattern=[
                    "blocks.{block}.attn.norm_q.txt.weight",
                    "blocks.{block}.attn.norm_q.all.weight",
                ],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.norm_k_vid.weight",
                from_pattern=[
                    "blocks.{block}.attn.norm_k.vid.weight",
                    "blocks.{block}.attn.norm_k.all.weight",
                ],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.norm_k_txt.weight",
                from_pattern=[
                    "blocks.{block}.attn.norm_k.txt.weight",
                    "blocks.{block}.attn.norm_k.all.weight",
                ],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.proj_out_vid.weight",
                from_pattern=[
                    "blocks.{block}.attn.proj_out.vid.weight",
                    "blocks.{block}.attn.proj_out.all.weight",
                ],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.proj_out_vid.bias",
                from_pattern=[
                    "blocks.{block}.attn.proj_out.vid.bias",
                    "blocks.{block}.attn.proj_out.all.bias",
                ],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.proj_out_txt.weight",
                from_pattern=[
                    "blocks.{block}.attn.proj_out.txt.weight",
                    "blocks.{block}.attn.proj_out.all.weight",
                ],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.proj_out_txt.bias",
                from_pattern=[
                    "blocks.{block}.attn.proj_out.txt.bias",
                    "blocks.{block}.attn.proj_out.all.bias",
                ],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.rope.freqs",
                from_pattern=["blocks.{block}.attn.rope.rope.freqs"],
                max_blocks=num_blocks,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.mlp.vid.proj_in.weight",
                from_pattern=["blocks.{block}.mlp.vid.proj_in.weight"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.mlp.vid.proj_in_gate.weight",
                from_pattern=["blocks.{block}.mlp.vid.proj_in_gate.weight"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.mlp.vid.proj_out.weight",
                from_pattern=["blocks.{block}.mlp.vid.proj_out.weight"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.mlp.txt.proj_in.weight",
                from_pattern=["blocks.{block}.mlp.txt.proj_in.weight"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.mlp.txt.proj_in_gate.weight",
                from_pattern=["blocks.{block}.mlp.txt.proj_in_gate.weight"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.mlp.txt.proj_out.weight",
                from_pattern=["blocks.{block}.mlp.txt.proj_out.weight"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.mlp.all.proj_in.weight",
                from_pattern=["blocks.{block}.mlp.all.proj_in.weight"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.mlp.all.proj_in_gate.weight",
                from_pattern=["blocks.{block}.mlp.all.proj_in_gate.weight"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.mlp.all.proj_out.weight",
                from_pattern=["blocks.{block}.mlp.all.proj_out.weight"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_vid.attn_shift",
                from_pattern=["blocks.{block}.ada.vid.attn_shift"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_vid.attn_scale",
                from_pattern=["blocks.{block}.ada.vid.attn_scale"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_vid.attn_gate",
                from_pattern=["blocks.{block}.ada.vid.attn_gate"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_vid.mlp_shift",
                from_pattern=["blocks.{block}.ada.vid.mlp_shift"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_vid.mlp_scale",
                from_pattern=["blocks.{block}.ada.vid.mlp_scale"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_vid.mlp_gate",
                from_pattern=["blocks.{block}.ada.vid.mlp_gate"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_txt.attn_shift",
                from_pattern=["blocks.{block}.ada.txt.attn_shift"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_txt.attn_scale",
                from_pattern=["blocks.{block}.ada.txt.attn_scale"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_txt.attn_gate",
                from_pattern=["blocks.{block}.ada.txt.attn_gate"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_txt.mlp_shift",
                from_pattern=["blocks.{block}.ada.txt.mlp_shift"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_txt.mlp_scale",
                from_pattern=["blocks.{block}.ada.txt.mlp_scale"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_txt.mlp_gate",
                from_pattern=["blocks.{block}.ada.txt.mlp_gate"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_all.attn_shift",
                from_pattern=["blocks.{block}.ada.all.attn_shift"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_all.attn_scale",
                from_pattern=["blocks.{block}.ada.all.attn_scale"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_all.attn_gate",
                from_pattern=["blocks.{block}.ada.all.attn_gate"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_all.mlp_shift",
                from_pattern=["blocks.{block}.ada.all.mlp_shift"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_all.mlp_scale",
                from_pattern=["blocks.{block}.ada.all.mlp_scale"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="blocks.{block}.ada.params_all.mlp_gate",
                from_pattern=["blocks.{block}.ada.all.mlp_gate"],
                max_blocks=num_blocks,
                required=False,
            ),
            WeightTarget(
                to_pattern="vid_out_norm.weight",
                from_pattern=["vid_out_norm.weight"],
            ),
            WeightTarget(
                to_pattern="out_shift",
                from_pattern=["vid_out_ada.out_shift"],
            ),
            WeightTarget(
                to_pattern="out_scale",
                from_pattern=["vid_out_ada.out_scale"],
            ),
            WeightTarget(
                to_pattern="vid_out.proj.weight",
                from_pattern=["vid_out.proj.weight"],
            ),
            WeightTarget(
                to_pattern="vid_out.proj.bias",
                from_pattern=["vid_out.proj.bias"],
            ),
        ]

    @staticmethod
    def get_vae_mapping() -> List[WeightTarget]:
        return [
            WeightTarget(
                to_pattern="encoder.conv_in.weight",
                from_pattern=["encoder.conv_in.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.conv_in.bias",
                from_pattern=["encoder.conv_in.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.conv1.weight",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.conv1.bias",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.conv2.weight",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.conv2.bias",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.norm1.weight",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.norm1.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.norm1.bias",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.norm1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.norm2.weight",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.norm2.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.norm2.bias",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.norm2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.conv_shortcut.weight",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.conv_shortcut.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.conv_shortcut.bias",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.conv_shortcut.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.downsamplers.0.conv.weight",
                from_pattern=["encoder.down_blocks.{block}.downsamplers.0.conv.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.downsamplers.0.conv.bias",
                from_pattern=["encoder.down_blocks.{block}.downsamplers.0.conv.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.conv1.weight",
                from_pattern=["encoder.mid_block.resnets.{i}.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.conv1.bias",
                from_pattern=["encoder.mid_block.resnets.{i}.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.conv2.weight",
                from_pattern=["encoder.mid_block.resnets.{i}.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.conv2.bias",
                from_pattern=["encoder.mid_block.resnets.{i}.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.norm1.weight",
                from_pattern=["encoder.mid_block.resnets.{i}.norm1.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.norm1.bias",
                from_pattern=["encoder.mid_block.resnets.{i}.norm1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.norm2.weight",
                from_pattern=["encoder.mid_block.resnets.{i}.norm2.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.norm2.bias",
                from_pattern=["encoder.mid_block.resnets.{i}.norm2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.group_norm.weight",
                from_pattern=["encoder.mid_block.attentions.0.group_norm.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.group_norm.bias",
                from_pattern=["encoder.mid_block.attentions.0.group_norm.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_q.weight",
                from_pattern=["encoder.mid_block.attentions.0.to_q.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_q.bias",
                from_pattern=["encoder.mid_block.attentions.0.to_q.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_k.weight",
                from_pattern=["encoder.mid_block.attentions.0.to_k.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_k.bias",
                from_pattern=["encoder.mid_block.attentions.0.to_k.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_v.weight",
                from_pattern=["encoder.mid_block.attentions.0.to_v.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_v.bias",
                from_pattern=["encoder.mid_block.attentions.0.to_v.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_out.0.weight",
                from_pattern=["encoder.mid_block.attentions.0.to_out.0.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_out.0.bias",
                from_pattern=["encoder.mid_block.attentions.0.to_out.0.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.conv_norm_out.weight",
                from_pattern=["encoder.conv_norm_out.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.conv_norm_out.bias",
                from_pattern=["encoder.conv_norm_out.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.conv_out.weight",
                from_pattern=["encoder.conv_out.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.conv_out.bias",
                from_pattern=["encoder.conv_out.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.conv_in.weight",
                from_pattern=["decoder.conv_in.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.conv_in.bias",
                from_pattern=["decoder.conv_in.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.conv1.weight",
                from_pattern=["decoder.mid_block.resnets.{i}.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.conv1.bias",
                from_pattern=["decoder.mid_block.resnets.{i}.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.conv2.weight",
                from_pattern=["decoder.mid_block.resnets.{i}.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.conv2.bias",
                from_pattern=["decoder.mid_block.resnets.{i}.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.norm1.weight",
                from_pattern=["decoder.mid_block.resnets.{i}.norm1.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.norm1.bias",
                from_pattern=["decoder.mid_block.resnets.{i}.norm1.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.norm2.weight",
                from_pattern=["decoder.mid_block.resnets.{i}.norm2.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.norm2.bias",
                from_pattern=["decoder.mid_block.resnets.{i}.norm2.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.group_norm.weight",
                from_pattern=["decoder.mid_block.attentions.0.group_norm.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.group_norm.bias",
                from_pattern=["decoder.mid_block.attentions.0.group_norm.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_q.weight",
                from_pattern=["decoder.mid_block.attentions.0.to_q.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_q.bias",
                from_pattern=["decoder.mid_block.attentions.0.to_q.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_k.weight",
                from_pattern=["decoder.mid_block.attentions.0.to_k.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_k.bias",
                from_pattern=["decoder.mid_block.attentions.0.to_k.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_v.weight",
                from_pattern=["decoder.mid_block.attentions.0.to_v.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_v.bias",
                from_pattern=["decoder.mid_block.attentions.0.to_v.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_out.0.weight",
                from_pattern=["decoder.mid_block.attentions.0.to_out.0.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_out.0.bias",
                from_pattern=["decoder.mid_block.attentions.0.to_out.0.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.conv1.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.conv1.bias",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.conv2.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.conv2.bias",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.norm1.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.norm1.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.norm1.bias",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.norm1.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.norm2.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.norm2.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.norm2.bias",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.norm2.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.conv_shortcut.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv_shortcut.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
                required=False,
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.conv_shortcut.bias",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv_shortcut.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.upsamplers.0.conv.weight",
                from_pattern=["decoder.up_blocks.{block}.upsamplers.0.conv.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
                required=False,
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.upsamplers.0.conv.bias",
                from_pattern=["decoder.up_blocks.{block}.upsamplers.0.conv.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.upsamplers.0.upscale_conv.weight",
                from_pattern=["decoder.up_blocks.{block}.upsamplers.0.upscale_conv.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
                required=False,
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.upsamplers.0.upscale_conv.bias",
                from_pattern=["decoder.up_blocks.{block}.upsamplers.0.upscale_conv.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="decoder.conv_norm_out.weight",
                from_pattern=["decoder.conv_norm_out.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.conv_norm_out.bias",
                from_pattern=["decoder.conv_norm_out.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.conv_out.weight",
                from_pattern=["decoder.conv_out.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.conv_out.bias",
                from_pattern=["decoder.conv_out.bias"],
            ),
        ]



import math

import mlx.core as mx
from mlx import nn


class QwenTimesteps(nn.Module):
    def __init__(self, proj_dim: int = 256, scale: float = 1000.0):
        super().__init__()
        self.proj_dim = proj_dim
        self.flip_sin_to_cos = True
        self.downscale_freq_shift = 0.0
        self.scale = scale

    def __call__(self, timesteps: mx.array) -> mx.array:
        assert len(timesteps.shape) == 1, "Timesteps should be a 1d-array"

        half_dim = self.proj_dim // 2
        max_period = 10000

        exponent = -math.log(max_period) * mx.arange(0, half_dim, dtype=mx.float32)
        exponent = exponent / (half_dim - self.downscale_freq_shift)

        timesteps_dtype = timesteps.dtype
        emb = mx.exp(exponent).astype(timesteps_dtype)

        emb = timesteps[:, None].astype(mx.float32) * emb[None, :]
        emb = self.scale * emb
        emb = mx.concatenate([mx.sin(emb), mx.cos(emb)], axis=-1)

        if self.flip_sin_to_cos:
            emb = mx.concatenate([emb[:, half_dim:], emb[:, :half_dim]], axis=-1)

        return emb



from pathlib import Path

from mflux.callbacks.callback_manager import CallbackManager
from mflux.cli.defaults import defaults as ui_defaults
from mflux.cli.parser.parsers import CommandLineParser
from mflux.models.flux.latent_creator.flux_latent_creator import FluxLatentCreator
from mflux.models.flux.variants.kontext.flux_kontext import Flux1Kontext
from mflux.utils.dimension_resolver import DimensionResolver
from mflux.utils.exceptions import PromptFileReadError, StopImageGenerationException
from mflux.utils.prompt_util import PromptUtil


def main():
    # 0. Parse command line arguments
    parser = CommandLineParser(description="Generate an image using Flux Kontext with image conditioning.")
    parser.add_general_arguments()
    parser.add_model_arguments(require_model_arg=False)
    parser.add_lora_arguments()
    parser.add_image_generator_arguments(supports_metadata_config=True, supports_dimension_scale_factor=True)
    parser.add_image_to_image_arguments(required=True)
    parser.add_output_arguments()
    args = parser.parse_args()

    # 0. Set default guidance value if not provided by user
    if args.guidance is None:
        args.guidance = ui_defaults.GUIDANCE_SCALE_KONTEXT

    # 1. Load the model
    flux = Flux1Kontext(
        quantize=args.quantize,
        model_path=args.model_path,
        lora_paths=args.lora_paths,
        lora_scales=args.lora_scales,
    )

    # 2. Register callbacks
    memory_saver = CallbackManager.register_callbacks(
        args=args,
        model=flux,
        latent_creator=FluxLatentCreator,
    )

    try:
        # Resolve dimensions (supports ScaleFactor like "2x" based on --image-path)
        width, height = DimensionResolver.resolve(
            width=args.width,
            height=args.height,
            reference_image_path=args.image_path,
        )

        for seed in args.seed:
            # 3. Generate an image for each seed value
            image = flux.generate_image(
                seed=seed,
                prompt=PromptUtil.read_prompt(args),
                width=width,
                height=height,
                guidance=args.guidance,
                scheduler=args.scheduler,
                image_path=args.image_path,
                num_inference_steps=args.steps,
            )

            # 4. Save the image
            output_path = Path(args.output.format(seed=seed))
            image.save(path=output_path, export_json_metadata=args.metadata)

    except (StopImageGenerationException, PromptFileReadError) as exc:
        print(exc)
    finally:
        if memory_saver:
            print(memory_saver.memory_stats())


if __name__ == "__main__":
    main()



import math
from pathlib import Path

import mlx.core as mx
from mlx import nn
from tqdm import tqdm

from mflux.models.common.config.config import Config
from mflux.models.common.config.model_config import ModelConfig
from mflux.models.common.vae.vae_util import VAEUtil
from mflux.models.qwen.latent_creator.qwen_latent_creator import QwenLatentCreator
from mflux.models.qwen.model.qwen_text_encoder.qwen_text_encoder import QwenTextEncoder
from mflux.models.qwen.model.qwen_transformer.qwen_transformer import QwenTransformer
from mflux.models.qwen.model.qwen_vae.qwen_vae import QwenVAE
from mflux.models.qwen.qwen_initializer import QwenImageInitializer
from mflux.models.qwen.variants.edit.qwen_edit_util import QwenEditUtil
from mflux.models.qwen.variants.txt2img.qwen_image import QwenImage
from mflux.utils.exceptions import StopImageGenerationException
from mflux.utils.generated_image import GeneratedImage
from mflux.utils.image_util import ImageUtil


class QwenImageEdit(nn.Module):
    vae: QwenVAE
    transformer: QwenTransformer
    text_encoder: QwenTextEncoder

    def __init__(
        self,
        quantize: int | None = None,
        model_path: str | None = None,
        lora_paths: list[str] | None = None,
        lora_scales: list[float] | None = None,
        model_config: ModelConfig = ModelConfig.qwen_image_edit(),
    ):
        super().__init__()
        QwenImageInitializer.init_edit(
            model=self,
            quantize=quantize,
            model_path=model_path,
            lora_paths=lora_paths,
            lora_scales=lora_scales,
            model_config=model_config,
        )

    def generate_image(
        self,
        seed: int,
        prompt: str,
        image_paths: list[str],
        num_inference_steps: int = 4,
        height: int | None = None,
        width: int | None = None,
        guidance: float = 4.0,
        image_path: Path | str | None = None,
        scheduler: str = "linear",
        negative_prompt: str | None = None,
    ) -> GeneratedImage:
        config, vl_width, vl_height, vae_width, vae_height = self._compute_dimensions(
            width=width,
            height=height,
            guidance=guidance,
            scheduler=scheduler,
            image_path=image_path,
            image_paths=image_paths,
            num_inference_steps=num_inference_steps,
        )
        timesteps = config.scheduler.timesteps
        time_steps = tqdm(range(len(timesteps)))

        # 1. Create initial latents
        latents = QwenLatentCreator.create_noise(
            seed=seed,
            width=config.width,
            height=config.height,
        )

        # 2. Encode the prompt
        prompt_embeds, prompt_mask, negative_prompt_embeds, negative_prompt_mask = self._encode_prompts_with_images(
            prompt=prompt,
            config=config,
            vl_width=vl_width,
            vl_height=vl_height,
            image_paths=image_paths,
            negative_prompt=negative_prompt,
        )

        # 3. Generate image conditioning latents
        static_image_latents, qwen_image_ids, cond_h_patches, cond_w_patches, num_images = (
            QwenEditUtil.create_image_conditioning_latents(
                vae=self.vae,
                width=vae_width,
                height=vae_height,
                vl_width=vl_width,
                vl_height=vl_height,
                image_paths=image_paths,
                tiling_config=self.tiling_config,
            )
        )

        # 4. Create callback context and call before_loop
        ctx = self.callbacks.start(seed=seed, prompt=prompt, config=config)
        ctx.before_loop(latents)

        for t in time_steps:
            try:
                # 5.t Concatenate the updated latents with the static image latents
                hidden_states = mx.concatenate([latents, static_image_latents], axis=1)
                hidden_states_neg = mx.concatenate([latents, static_image_latents], axis=1)

                # 6.t Predict the noise
                if num_images > 1:
                    cond_image_grid = [(1, cond_h_patches, cond_w_patches) for _ in range(num_images)]
                else:
                    cond_image_grid = (1, cond_h_patches, cond_w_patches)

                noise = self.transformer(
                    t=t,
                    config=config,
                    hidden_states=hidden_states,
                    encoder_hidden_states=prompt_embeds,
                    encoder_hidden_states_mask=prompt_mask,
                    qwen_image_ids=qwen_image_ids,
                    cond_image_grid=cond_image_grid,
                )[:, : latents.shape[1]]
                noise_negative = self.transformer(
                    t=t,
                    config=config,
                    hidden_states=hidden_states_neg,
                    encoder_hidden_states=negative_prompt_embeds,
                    encoder_hidden_states_mask=negative_prompt_mask,
                    qwen_image_ids=qwen_image_ids,
                    cond_image_grid=cond_image_grid,
                )[:, : latents.shape[1]]
                guided_noise = QwenImage.compute_guided_noise(noise, noise_negative, config.guidance)

                # 7.t Take one denoise step
                latents = config.scheduler.step(noise=guided_noise, timestep=t, latents=latents)

                # 8.t Call subscribers in-loop
                ctx.in_loop(t, latents, time_steps=time_steps)

                # (Optional) Evaluate to enable progress tracking
                mx.eval(latents)

            except KeyboardInterrupt:  # noqa: PERF203
                ctx.interruption(t, latents, time_steps=time_steps)
                raise StopImageGenerationException(f"Stopping image generation at step {t + 1}/{len(timesteps)}")

        # 9. Call subscribers after loop
        ctx.after_loop(latents)

        # 10. Decode the latent array and return the image
        latents = QwenLatentCreator.unpack_latents(latents=latents, height=config.height, width=config.width)
        decoded = VAEUtil.decode(vae=self.vae, latent=latents, tiling_config=self.tiling_config)
        return ImageUtil.to_image(
            decoded_latents=decoded,
            config=config,
            seed=seed,
            prompt=prompt,
            quantization=self.bits,
            lora_paths=self.lora_paths,
            lora_scales=self.lora_scales,
            image_path=config.image_path,
            image_paths=image_paths,
            generation_time=time_steps.format_dict["elapsed"],
            negative_prompt=negative_prompt,
        )

    def _encode_prompts_with_images(
        self,
        prompt: str,
        negative_prompt: str,
        image_paths: list[str],
        config,
        vl_width: int | None = None,
        vl_height: int | None = None,
    ) -> tuple[mx.array, mx.array, mx.array, mx.array]:
        tokenizer = self.tokenizers["qwen_vl"]
        pos_input_ids, pos_attention_mask, pos_pixel_values, pos_image_grid_thw = tokenizer.tokenize_with_image(
            prompt, image_paths, vl_width=vl_width, vl_height=vl_height
        )

        pos_hidden_states = self.qwen_vl_encoder(
            input_ids=pos_input_ids,
            attention_mask=pos_attention_mask,
            pixel_values=pos_pixel_values,
            image_grid_thw=pos_image_grid_thw,
        )

        neg_prompt = negative_prompt if negative_prompt is not None else ""

        # Use real MLX tokenizer for negative prompt
        neg_input_ids, neg_attention_mask, neg_pixel_values, neg_image_grid_thw = tokenizer.tokenize_with_image(
            neg_prompt, image_paths, vl_width=vl_width, vl_height=vl_height
        )

        neg_hidden_states = self.qwen_vl_encoder(
            input_ids=neg_input_ids,
            attention_mask=neg_attention_mask,
            pixel_values=neg_pixel_values,
            image_grid_thw=neg_image_grid_thw,
        )

        final_prompt_embeds = pos_hidden_states[0].astype(mx.float16)
        final_prompt_mask = pos_hidden_states[1].astype(mx.float16)

        return (
            final_prompt_embeds,  # prompt_embeds
            final_prompt_mask,  # prompt_mask
            neg_hidden_states[0].astype(mx.float16),  # negative_prompt_embeds
            neg_hidden_states[1].astype(mx.float16),  # negative_prompt_mask
        )

    def _compute_dimensions(
        self,
        image_paths: list[str],
        num_inference_steps: int,
        height: int | None,
        width: int | None,
        guidance: float,
        image_path: Path | str | None,
        scheduler: str,
    ) -> tuple[Config, int, int, int, int]:
        last_image = ImageUtil.load_image(image_paths[-1]).convert("RGB")
        image_size = last_image.size

        target_area = 1024 * 1024
        ratio = image_size[0] / image_size[1]
        calculated_width = math.sqrt(target_area * ratio)
        calculated_height = calculated_width / ratio
        calculated_width = round(calculated_width / 32) * 32
        calculated_height = round(calculated_height / 32) * 32

        use_height = height or int(calculated_height)
        use_width = width or int(calculated_width)

        vae_scale_factor = 8
        multiple_of = vae_scale_factor * 2
        use_width = use_width // multiple_of * multiple_of
        use_height = use_height // multiple_of * multiple_of

        config = Config(
            width=use_width,
            height=use_height,
            guidance=guidance,
            scheduler=scheduler,
            image_path=image_path,
            model_config=self.model_config,
            num_inference_steps=num_inference_steps,
        )

        CONDITION_IMAGE_SIZE = 384 * 384
        condition_ratio = image_size[0] / image_size[1]
        vl_width = math.sqrt(CONDITION_IMAGE_SIZE * condition_ratio)
        vl_height = vl_width / condition_ratio
        vl_width = round(vl_width / 32) * 32
        vl_height = round(vl_height / 32) * 32

        VAE_IMAGE_SIZE = 1024 * 1024
        vae_ratio = image_size[0] / image_size[1]
        vae_width = math.sqrt(VAE_IMAGE_SIZE * vae_ratio)
        vae_height = vae_width / vae_ratio
        vae_width = round(vae_width / 32) * 32
        vae_height = round(vae_height / 32) * 32

        return config, int(vl_width), int(vl_height), int(vae_width), int(vae_height)



import mlx.core as mx
from mlx import nn

from mflux.models.qwen.model.qwen_vae.qwen_image_causal_conv_3d import QwenImageCausalConv3D
from mflux.models.qwen.model.qwen_vae.qwen_image_down_block_3d import QwenImageDownBlock3D
from mflux.models.qwen.model.qwen_vae.qwen_image_mid_block_3d import QwenImageMidBlock3D
from mflux.models.qwen.model.qwen_vae.qwen_image_rms_norm import QwenImageRMSNorm


class QwenImageEncoder3D(nn.Module):
    def __init__(self):
        super().__init__()
        self.dim = 96
        self.z_dim = 32
        self.dim_mult = [1, 2, 4, 4]
        self.num_res_blocks = [2, 2, 2, 2]
        self.attn_scales = []
        self.temporal_downsample = [False, False, True, True]
        self.dropout = 0.0

        dims = [self.dim * u for u in [1] + self.dim_mult]
        self.conv_in = QwenImageCausalConv3D(3, dims[0], 3, 1, 1)

        down_blocks = []
        for i, (in_dim, out_dim) in enumerate(zip(dims[:-1], dims[1:])):
            downsample_mode = "downsample3d" if self.temporal_downsample[i] else "downsample2d"
            # Do not downsample on the final stage only
            if i == len(dims) - 2:
                downsample_mode = None
            # Use per-stage num_res_blocks
            stage_res_blocks = self.num_res_blocks[i] if isinstance(self.num_res_blocks, list) else self.num_res_blocks
            down_block = QwenImageDownBlock3D(
                in_dim, out_dim, num_res_blocks=stage_res_blocks, downsample_mode=downsample_mode
            )
            down_blocks.append(down_block)
        self.down_blocks = down_blocks

        self.mid_block = QwenImageMidBlock3D(dims[-1], num_layers=1)
        self.norm_out = QwenImageRMSNorm(dims[-1], images=False)
        self.conv_out = QwenImageCausalConv3D(dims[-1], 32, 3, 1, 1)

    def __call__(self, x: mx.array) -> mx.array:
        x = self.conv_in(x)
        for stage_idx, down_block in enumerate(self.down_blocks):
            if stage_idx == 3:
                for resnet in down_block.resnets:
                    residual = x
                    n1 = resnet.norm1(x)
                    a1 = nn.silu(n1)
                    c1 = resnet.conv1(a1)
                    n2 = resnet.norm2(c1)
                    a2 = nn.silu(n2)
                    c2 = resnet.conv2(a2)
                    x = c2 + residual
                if down_block.downsamplers is not None:
                    x = down_block.downsamplers[0](x)
            else:
                x = down_block(x)

        x = self.mid_block(x)
        norm_in = x
        x = self.norm_out(norm_in)
        x = nn.silu(x)
        encoded = self.conv_out(x)
        return encoded



import mlx.core as mx
from mlx import nn

from mflux.models.fibo_vlm.model.qwen3_vl_decoder_layer import Qwen3VLDecoderLayer
from mflux.models.fibo_vlm.model.qwen3_vl_rms_norm import Qwen3VLRMSNorm
from mflux.models.fibo_vlm.model.qwen3_vl_rope import Qwen3VLRotaryEmbedding


class Qwen3VLDecoder(nn.Module):
    def __init__(
        self,
        vocab_size: int = 151936,
        hidden_size: int = 2560,
        num_hidden_layers: int = 36,
        num_attention_heads: int = 32,
        num_key_value_heads: int = 8,
        intermediate_size: int = 9728,
        max_position_embeddings: int = 262144,
        rope_theta: float = 5000000.0,
        rms_norm_eps: float = 1e-6,
        head_dim: int = 128,
        attention_bias: bool = False,
        mrope_section: list[int] | None = None,
        attention_scaling: float = 1.0,
        image_token_id: int = 151655,
        visual=None,
    ):
        super().__init__()
        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.num_hidden_layers = num_hidden_layers
        if mrope_section is None:
            mrope_section = [24, 20, 20]

        self.embed_tokens = nn.Embedding(vocab_size, hidden_size)
        self.layers = [
            Qwen3VLDecoderLayer(
                hidden_size=hidden_size,
                num_attention_heads=num_attention_heads,
                num_key_value_heads=num_key_value_heads,
                head_dim=head_dim,
                max_position_embeddings=max_position_embeddings,
                rope_theta=rope_theta,
                mrope_section=mrope_section,
                attention_bias=attention_bias,
                rms_norm_eps=rms_norm_eps,
                intermediate_size=intermediate_size,
            )
            for _ in range(num_hidden_layers)
        ]
        self.norm = Qwen3VLRMSNorm(hidden_size, eps=rms_norm_eps)
        self.rotary_emb = Qwen3VLRotaryEmbedding(
            dim=head_dim,
            max_position_embeddings=max_position_embeddings,
            base=rope_theta,
            scaling_factor=attention_scaling,
            mrope_section=mrope_section,
        )
        self.lm_head = nn.Linear(hidden_size, vocab_size, bias=False)
        self.visual = visual
        self.image_token_id = image_token_id

    def __call__(
        self,
        input_ids: mx.array,
        attention_mask: mx.array | None = None,
        pixel_values: mx.array | None = None,
        image_grid_thw: mx.array | None = None,
        use_cache: bool = False,
        past_key_values: list[tuple[mx.array, mx.array]] | None = None,
    ) -> mx.array | tuple[mx.array, list[tuple[mx.array, mx.array]]]:
        batch_size, seq_len = input_ids.shape

        # Get embeddings
        inputs_embeds = self.embed_tokens(input_ids)

        # Handle vision inputs if provided
        if pixel_values is not None and image_grid_thw is not None and self.visual is not None:
            image_embeds_split = self._get_image_features(pixel_values, image_grid_thw)
            image_embeds = mx.concatenate(image_embeds_split, axis=0)

            if self.image_token_id is not None:
                image_positions = input_ids == self.image_token_id
                n_image_tokens = mx.sum(image_positions).item()

                if n_image_tokens > 0 and image_embeds.shape[0] >= n_image_tokens:
                    image_positions_flat = image_positions.flatten()
                    inputs_embeds_flat = inputs_embeds.reshape(-1, inputs_embeds.shape[-1])

                    new_embeds_list = []
                    image_idx = 0
                    for i in range(len(image_positions_flat)):
                        if image_positions_flat[i] and image_idx < image_embeds.shape[0]:
                            new_embeds_list.append(image_embeds[image_idx])
                            image_idx += 1
                        else:
                            new_embeds_list.append(inputs_embeds_flat[i])

                    new_embeds = mx.stack(new_embeds_list, axis=0)
                    inputs_embeds = new_embeds.reshape(inputs_embeds.shape)

        # Create attention mask with causal masking (from QwenEncoder)
        if attention_mask is None:
            attention_mask = mx.ones((batch_size, seq_len), dtype=mx.int32)

        # Position embeddings
        if use_cache and past_key_values is not None:
            cached_seq_len = past_key_values[0][0].shape[2] if len(past_key_values) > 0 else 0
            cache_position = mx.arange(cached_seq_len, cached_seq_len + seq_len, dtype=mx.int32)
        else:
            cache_position = mx.arange(seq_len, dtype=mx.int32)
        position_ids = mx.expand_dims(mx.expand_dims(cache_position, axis=0), axis=0)
        position_ids = mx.broadcast_to(position_ids, (3, batch_size, seq_len))

        # Create causal + padding mask
        mask_dtype = inputs_embeds.dtype
        padding_mask = mx.where(
            attention_mask == 1,
            mx.zeros(attention_mask.shape, dtype=mask_dtype),
            mx.full(attention_mask.shape, -float("inf"), dtype=mask_dtype),
        )
        padding_mask = mx.expand_dims(mx.expand_dims(padding_mask, axis=1), axis=1)

        # Optimize: when seq_len=1 (cached iterations), causal mask is trivial (all zeros)
        if seq_len == 1:
            causal_tri_mask = mx.zeros((batch_size, 1, 1, 1), dtype=mask_dtype)
        else:
            idx = mx.arange(seq_len, dtype=mx.int32)
            j = mx.expand_dims(idx, axis=0)
            i = mx.expand_dims(idx, axis=1)
            tri_bool = j > i
            zeros_2d = mx.zeros((seq_len, seq_len), dtype=mask_dtype)
            neginf_2d = mx.full((seq_len, seq_len), -float("inf"), dtype=mask_dtype)
            causal_tri_mask = mx.where(tri_bool, neginf_2d, zeros_2d)
            causal_tri_mask = mx.expand_dims(mx.expand_dims(causal_tri_mask, axis=0), axis=0)
            causal_tri_mask = mx.broadcast_to(causal_tri_mask, (batch_size, 1, seq_len, seq_len))
        attention_mask_4d = causal_tri_mask + padding_mask
        hidden_states = inputs_embeds
        position_embeddings = self.rotary_emb(hidden_states, position_ids)

        present_key_values = [] if use_cache else None
        for layer_idx, layer in enumerate(self.layers):
            layer_past = None
            if use_cache and past_key_values is not None:
                layer_past = past_key_values[layer_idx] if layer_idx < len(past_key_values) else None

            layer_output = layer(
                hidden_states,
                attention_mask_4d,
                position_embeddings,
                past_key_value=layer_past,
            )

            if use_cache:
                hidden_states, layer_present = layer_output
                present_key_values.append(layer_present)
            else:
                hidden_states = layer_output

        # Final norm
        hidden_states = self.norm(hidden_states)
        hidden_states_f32 = hidden_states.astype(mx.float32)
        weight_f32 = self.lm_head.weight.astype(mx.float32)
        logits_f32 = mx.matmul(hidden_states_f32, weight_f32.T)
        logits = logits_f32.astype(hidden_states.dtype)

        if use_cache:
            return logits, present_key_values
        return logits

    def _get_image_features(self, pixel_values: mx.array, image_grid_thw: mx.array) -> list[mx.array]:
        if self.visual is None:
            raise RuntimeError("Vision transformer not initialized. Call load_visual_weights() first.")

        # Vision model can handle the original dtype (usually float16)
        image_embeds, _ = self.visual(pixel_values, image_grid_thw)  # Ignore deepstack for now
        original_split_sizes = image_grid_thw.prod(axis=-1).astype(mx.int32)
        spatial_merge_size = self.visual.spatial_merge_size
        split_sizes = (original_split_sizes // (spatial_merge_size**2)).astype(mx.int32)
        split_sizes = [int(s) for s in split_sizes.tolist()]
        split_sizes = [s for s in split_sizes if s > 0]

        image_embeds_split = []
        start_idx = 0
        for split_size in split_sizes:
            end_idx = start_idx + split_size
            image_embeds_split.append(image_embeds[start_idx:end_idx])
            start_idx = end_idx
        return image_embeds_split



import mlx.core as mx
from mlx import nn


class T5LayerNorm(nn.Module):
    def __init__(self):
        super().__init__()
        self.weight = mx.ones((4096,))
        self.variance_epsilon = 1e-06

    def __call__(self, hidden_states: mx.array) -> mx.array:
        variance = mx.mean(
            mx.power(hidden_states.astype(mx.float32), 2),
            axis=-1,
            keepdims=True,
        )
        hidden_states = hidden_states * mx.rsqrt(variance + self.variance_epsilon)
        return self.weight * hidden_states



import mlx.core as mx
from mlx import nn

from mflux.models.qwen.model.qwen_vae.qwen_image_causal_conv_3d import QwenImageCausalConv3D


class QwenImageResample3D(nn.Module):
    def __init__(self, dim: int, mode: str):
        super().__init__()
        self.dim = dim
        self.mode = mode
        if mode == "upsample3d":
            self.time_conv = QwenImageCausalConv3D(dim, dim * 2, kernel_size=(3, 1, 1), stride=1, padding=(1, 0, 0))
            self.resample_conv = nn.Conv2d(dim, dim // 2, kernel_size=3, stride=1, padding=1)
        elif mode == "upsample2d":
            self.resample_conv = nn.Conv2d(dim, dim // 2, kernel_size=3, stride=1, padding=1)
        elif mode == "downsample3d":
            self.time_conv = QwenImageCausalConv3D(dim, dim, kernel_size=(3, 1, 1), stride=(2, 1, 1), padding=(0, 0, 0))
            self.resample_conv = nn.Conv2d(dim, dim, kernel_size=3, stride=2, padding=0)
        elif mode == "downsample2d":
            self.resample_conv = nn.Conv2d(dim, dim, kernel_size=3, stride=2, padding=0)
        else:
            raise ValueError(f"Unsupported resample mode: {mode}")

    def __call__(self, x: mx.array) -> mx.array:
        b, c, t, h, w = x.shape
        t = x.shape[2]
        x = mx.transpose(x, (0, 2, 1, 3, 4))
        x = mx.reshape(x, (b * t, c, h, w))
        x = mx.transpose(x, (0, 2, 3, 1))
        if self.mode in ["upsample3d", "upsample2d"]:
            x = self._up_sample_nearest_2x(x)

        if self.mode in ["downsample2d", "downsample3d"]:
            x = mx.pad(x, [(0, 0), (0, 1), (0, 1), (0, 0)])

        x = self.resample_conv(x)
        x = mx.transpose(x, (0, 3, 1, 2))
        new_c = x.shape[1]
        new_h, new_w = x.shape[2], x.shape[3]
        x = mx.reshape(x, (b, t, new_c, new_h, new_w))
        x = mx.transpose(x, (0, 2, 1, 3, 4))
        return x

    @staticmethod
    def _up_sample_nearest_2x(x: mx.array) -> mx.array:
        B_T, H, W, C = x.shape
        x = mx.repeat(x, 2, axis=1)
        x = mx.repeat(x, 2, axis=2)
        return x



import math
from typing import Tuple

import mlx.core as mx
from mlx import nn
from mlx.core.fast import scaled_dot_product_attention

from .smol_lm3_3b_rope import SmolLM3_3B_RotaryEmbedding


class SmolLM3_3B_SelfAttention(nn.Module):
    def __init__(
        self,
        hidden_size: int,
        num_attention_heads: int,
        num_key_value_heads: int,
        max_position_embeddings: int = 65_536,
        rope_theta: float = 5_000_000.0,
        attention_dropout: float = 0.0,
    ):
        super().__init__()
        self.hidden_size = hidden_size
        self.num_attention_heads = num_attention_heads
        self.num_key_value_heads = num_key_value_heads
        self.head_dim = hidden_size // num_attention_heads
        self.num_key_value_groups = num_attention_heads // num_key_value_heads
        self.scale = 1.0 / math.sqrt(self.head_dim)
        self.attention_dropout = attention_dropout

        # Projections – no bias, matching SmolLM3 config (attention_bias = False)
        self.q_proj = nn.Linear(hidden_size, num_attention_heads * self.head_dim, bias=False)
        self.k_proj = nn.Linear(hidden_size, num_key_value_heads * self.head_dim, bias=False)
        self.v_proj = nn.Linear(hidden_size, num_key_value_heads * self.head_dim, bias=False)
        self.o_proj = nn.Linear(num_attention_heads * self.head_dim, hidden_size, bias=False)

        self.rotary_emb = SmolLM3_3B_RotaryEmbedding(
            dim=self.head_dim,
            max_position_embeddings=max_position_embeddings,
            base=rope_theta,
        )

    def __call__(
        self,
        hidden_states: mx.array,
        attention_mask: mx.array | None = None,
        cos_sin: Tuple[mx.array, mx.array] | None = None,
    ) -> mx.array:
        batch_size, seq_len, _ = hidden_states.shape

        # Projections: Use nn.Linear directly (like Qwen/Flux models)
        q = self.q_proj(hidden_states)
        k = self.k_proj(hidden_states)
        v = self.v_proj(hidden_states)

        # Cast back to input dtype if Linear output is float32 (for consistency)
        if q.dtype != hidden_states.dtype:
            q = q.astype(hidden_states.dtype)
            k = k.astype(hidden_states.dtype)
            v = v.astype(hidden_states.dtype)

        # Reshape to (batch, heads, seq, head_dim)
        q = q.reshape(batch_size, seq_len, self.num_attention_heads, self.head_dim).transpose(0, 2, 1, 3)
        k = k.reshape(batch_size, seq_len, self.num_key_value_heads, self.head_dim).transpose(0, 2, 1, 3)
        v = v.reshape(batch_size, seq_len, self.num_key_value_heads, self.head_dim).transpose(0, 2, 1, 3)

        # Rotary embeddings
        if cos_sin is None:
            cos_sin = self.rotary_emb(seq_len)
        cos, sin = cos_sin
        q, k = SmolLM3_3B_SelfAttention._apply_rope(q, k, cos, sin)

        # Grouped-query attention: repeat kv heads if needed
        if self.num_key_value_heads != self.num_attention_heads:
            k = SmolLM3_3B_SelfAttention._repeat_kv(k, self.num_key_value_groups)
            v = SmolLM3_3B_SelfAttention._repeat_kv(v, self.num_key_value_groups)

        # Prepare mask for scaled_dot_product_attention
        attn_mask = None
        if attention_mask is not None:
            seq_len_k = k.shape[2]
            causal_mask = attention_mask[:, :, :, :seq_len_k]
            if causal_mask.shape[1] == 1:
                causal_mask = mx.broadcast_to(causal_mask, (batch_size, self.num_attention_heads, seq_len, seq_len_k))
            attn_mask = causal_mask.astype(q.dtype)

        # Use MLX fast SDPA
        attn_output = scaled_dot_product_attention(
            q,
            k,
            v,
            scale=self.scale,
            mask=attn_mask,
        )

        attn_output = attn_output.transpose(0, 2, 1, 3).reshape(batch_size, seq_len, self.hidden_size)
        attn_output = self.o_proj(attn_output)

        return attn_output

    @staticmethod
    def _rotate_half(x: mx.array) -> mx.array:
        x1 = x[..., : x.shape[-1] // 2]
        x2 = x[..., x.shape[-1] // 2 :]
        return mx.concatenate([-x2, x1], axis=-1)

    @staticmethod
    def _apply_rope(
        q: mx.array,
        k: mx.array,
        cos: mx.array,
        sin: mx.array,
    ) -> Tuple[mx.array, mx.array]:
        q_dtype = q.dtype
        k_dtype = k.dtype
        q = q.astype(mx.float32)
        k = k.astype(mx.float32)
        cos = cos.astype(mx.float32)
        sin = sin.astype(mx.float32)

        q_embed = (q * cos) + (SmolLM3_3B_SelfAttention._rotate_half(q) * sin)
        k_embed = (k * cos) + (SmolLM3_3B_SelfAttention._rotate_half(k) * sin)

        return q_embed.astype(q_dtype), k_embed.astype(k_dtype)

    @staticmethod
    def _repeat_kv(hidden_states: mx.array, n_rep: int) -> mx.array:
        batch, num_kv_heads, seq_len, head_dim = hidden_states.shape
        hidden_states = mx.expand_dims(hidden_states, axis=2)
        hidden_states = mx.broadcast_to(hidden_states, (batch, num_kv_heads, n_rep, seq_len, head_dim))
        return hidden_states.reshape(batch, num_kv_heads * n_rep, seq_len, head_dim)



import mlx.core as mx
from mlx import nn


class QwenMLP(nn.Module):
    def __init__(self, hidden_size: int, intermediate_size: int):
        super().__init__()
        self.gate_proj = nn.Linear(hidden_size, intermediate_size, bias=False)
        self.up_proj = nn.Linear(hidden_size, intermediate_size, bias=False)
        self.down_proj = nn.Linear(intermediate_size, hidden_size, bias=False)

    def __call__(self, hidden_states: mx.array) -> mx.array:
        gate_output = nn.silu(self.gate_proj(hidden_states))
        up_output = self.up_proj(hidden_states)
        intermediate_output = gate_output * up_output
        output = self.down_proj(intermediate_output)
        return output



import mlx.core as mx
from mlx import nn

from mflux.models.common.config.config import Config
from mflux.models.common.config.model_config import ModelConfig
from mflux.models.flux.model.flux_transformer.embed_nd import EmbedND
from mflux.models.flux.model.flux_transformer.joint_transformer_block import JointTransformerBlock
from mflux.models.flux.model.flux_transformer.single_transformer_block import SingleTransformerBlock
from mflux.models.flux.model.flux_transformer.time_text_embed import TimeTextEmbed
from mflux.models.flux.model.flux_transformer.transformer import Transformer


class TransformerControlnet(nn.Module):
    def __init__(
        self,
        model_config: ModelConfig,
        num_transformer_blocks: int = 5,
        num_single_transformer_blocks: int = 0,
    ):
        super().__init__()
        self.pos_embed = EmbedND()
        self.x_embedder = nn.Linear(64, 3072)
        self.time_text_embed = TimeTextEmbed(model_config=model_config)
        self.context_embedder = nn.Linear(4096, 3072)
        self.transformer_blocks = [JointTransformerBlock(i) for i in range(num_transformer_blocks)]
        self.single_transformer_blocks = [SingleTransformerBlock(i) for i in range(num_single_transformer_blocks)]

        self.controlnet_x_embedder = nn.Linear(64, 3072).apply(nn.init.constant(0))
        self.controlnet_blocks = [nn.Linear(3072, 3072).apply(nn.init.constant(0)) for _ in range(num_transformer_blocks)]  # fmt: off
        self.controlnet_single_blocks = [nn.Linear(3072, 3072) for _ in range(num_single_transformer_blocks)]

    def __call__(
        self,
        t: int,
        config: Config,
        hidden_states: mx.array,
        prompt_embeds: mx.array,
        pooled_prompt_embeds: mx.array,
        controlnet_condition: mx.array,
    ) -> tuple[list[mx.array], list[mx.array]]:
        # 1. Create embeddings
        hidden_states = self.x_embedder(hidden_states) + self.controlnet_x_embedder(controlnet_condition)
        encoder_hidden_states = self.context_embedder(prompt_embeds)
        text_embeddings = Transformer.compute_text_embeddings(t, pooled_prompt_embeds, self.time_text_embed, config)
        image_rotary_embeddings = Transformer.compute_rotary_embeddings(prompt_embeds, self.pos_embed, config, None)

        # 2. Run the joint flux_transformer blocks
        controlnet_block_samples = []
        for idx, block in enumerate(self.transformer_blocks):
            encoder_hidden_states, hidden_states = self._apply_joint_transformer_block(
                idx=idx,
                block=block,
                config=config,
                hidden_states=hidden_states,
                encoder_hidden_states=encoder_hidden_states,
                text_embeddings=text_embeddings,
                image_rotary_embeddings=image_rotary_embeddings,
                controlnet_block_samples=controlnet_block_samples,
            )

        # 3. Concat the hidden states
        hidden_states = mx.concatenate([encoder_hidden_states, hidden_states], axis=1)

        # 4. Run the single flux_transformer blocks
        controlnet_single_block_samples = []
        for idx, block in enumerate(self.single_transformer_blocks):
            hidden_states = self._apply_single_transformer_block(
                idx=idx,
                block=block,
                config=config,
                hidden_states=hidden_states,
                encoder_hidden_states=encoder_hidden_states,
                text_embeddings=text_embeddings,
                image_rotary_embeddings=image_rotary_embeddings,
                controlnet_single_block_samples=controlnet_single_block_samples,
            )

        return controlnet_block_samples, controlnet_single_block_samples

    def _apply_single_transformer_block(
        self,
        idx: int,
        config: Config,
        block: SingleTransformerBlock,
        hidden_states: mx.array,
        encoder_hidden_states: mx.array,
        text_embeddings: mx.array,
        image_rotary_embeddings: mx.array,
        controlnet_single_block_samples: list[mx.array],
    ) -> mx.array:
        # 1. Apply single flux_transformer block
        hidden_states = block(
            hidden_states=hidden_states,
            text_embeddings=text_embeddings,
            rotary_embeddings=image_rotary_embeddings,
        )

        # 2. Apply controlnet block
        states = hidden_states[:, encoder_hidden_states.shape[1] :]
        controlnet_sample = self.controlnet_single_blocks[idx](states)
        scaled_controlnet_sample = controlnet_sample * config.controlnet_strength
        controlnet_single_block_samples.append(scaled_controlnet_sample)

        return hidden_states

    def _apply_joint_transformer_block(
        self,
        idx: int,
        config: Config,
        block: JointTransformerBlock,
        hidden_states: mx.array,
        encoder_hidden_states: mx.array,
        text_embeddings: mx.array,
        image_rotary_embeddings: mx.array,
        controlnet_block_samples: list[mx.array],
    ) -> tuple[mx.array, mx.array]:
        # 1. Apply joint flux_transformer block
        encoder_hidden_states, hidden_states = block(
            hidden_states=hidden_states,
            encoder_hidden_states=encoder_hidden_states,
            text_embeddings=text_embeddings,
            rotary_embeddings=image_rotary_embeddings,
        )

        # 2. Apply controlnet block
        controlnet_sample = self.controlnet_blocks[idx](hidden_states)
        scaled_controlnet_example = controlnet_sample * config.controlnet_strength
        controlnet_block_samples.append(scaled_controlnet_example)

        return encoder_hidden_states, hidden_states



import mlx.core as mx
from mlx import nn

from mflux.models.common.lora.layer.linear_lora_layer import LoRALinear


class FusedLoRALinear(nn.Module):
    def __init__(self, base_linear: nn.Linear | nn.QuantizedLinear, loras: list[LoRALinear]):
        super().__init__()
        self.base_linear = base_linear
        self.loras = loras

    def __call__(self, x):
        base_out = self.base_linear(x)

        lora_out = mx.zeros_like(base_out)
        for lora in self.loras:
            lora_out += lora.scale * mx.matmul(mx.matmul(x, lora.lora_A), lora.lora_B)

        return base_out + lora_out



from __future__ import annotations

from typing import TYPE_CHECKING

from mflux.callbacks.callback import AfterLoopCallback, BeforeLoopCallback, InLoopCallback, InterruptCallback

if TYPE_CHECKING:
    from mflux.callbacks.generation_context import GenerationContext
    from mflux.models.common.config.config import Config


class CallbackRegistry:
    def __init__(self):
        self.in_loop: list[InLoopCallback] = []
        self.before_loop: list[BeforeLoopCallback] = []
        self.interrupt: list[InterruptCallback] = []
        self.after_loop: list[AfterLoopCallback] = []

    def register(self, callback) -> None:
        if hasattr(callback, "call_before_loop"):
            self.before_loop.append(callback)
        if hasattr(callback, "call_in_loop"):
            self.in_loop.append(callback)
        if hasattr(callback, "call_after_loop"):
            self.after_loop.append(callback)
        if hasattr(callback, "call_interrupt"):
            self.interrupt.append(callback)

    def start(self, seed: int, prompt: str, config: Config) -> GenerationContext:
        from mflux.callbacks.generation_context import GenerationContext

        return GenerationContext(self, seed, prompt, config)

    def before_loop_callbacks(self) -> list[BeforeLoopCallback]:
        return self.before_loop

    def in_loop_callbacks(self) -> list[InLoopCallback]:
        return self.in_loop

    def after_loop_callbacks(self) -> list[AfterLoopCallback]:
        return self.after_loop

    def interrupt_callbacks(self) -> list[InterruptCallback]:
        return self.interrupt



from pathlib import Path

import mlx.core as mx
from mlx import nn

from mflux.models.common.config.config import Config
from mflux.models.common.config.model_config import ModelConfig
from mflux.models.common.vae.vae_util import VAEUtil
from mflux.models.seedvr2.latent_creator.seedvr2_latent_creator import SeedVR2LatentCreator
from mflux.models.seedvr2.model.seedvr2_text_encoder.text_embeddings import SeedVR2TextEmbeddings
from mflux.models.seedvr2.model.seedvr2_transformer.transformer import SeedVR2Transformer
from mflux.models.seedvr2.model.seedvr2_vae.vae import SeedVR2VAE
from mflux.models.seedvr2.seedvr2_initializer import SeedVR2Initializer
from mflux.models.seedvr2.variants.upscale.seedvr2_util import SeedVR2Util
from mflux.utils.generated_image import GeneratedImage
from mflux.utils.image_util import ImageUtil
from mflux.utils.scale_factor import ScaleFactor


class SeedVR2(nn.Module):
    vae: SeedVR2VAE
    transformer: SeedVR2Transformer

    def __init__(
        self,
        quantize: int | None = None,
        model_path: str | None = None,
        model_config: ModelConfig = ModelConfig.seedvr2_3b(),
    ):
        super().__init__()
        SeedVR2Initializer.init(
            model=self,
            quantize=quantize,
            model_path=model_path,
            model_config=model_config,
        )

    def generate_image(
        self,
        seed: int,
        image_path: str | Path,
        resolution: int | ScaleFactor,
        softness: float = 0.0,
    ) -> GeneratedImage:
        # 0. Process and scale the input image
        processed_image, true_height, true_width = SeedVR2Util.preprocess_image(
            image_path=image_path,
            resolution=resolution,
            softness=softness,
        )

        # 1. Create a new config based on the model type and input parameters
        config = Config(
            width=true_width,
            height=true_height,
            guidance=1.0,
            num_inference_steps=1,
            image_path=image_path,
            scheduler="seedvr2_euler",
            model_config=self.model_config,
        )

        # 2. Create the initial latents and conditioning
        initial_latent = VAEUtil.encode(vae=self.vae, image=processed_image, tiling_config=self.tiling_config)
        static_condition = SeedVR2LatentCreator.create_condition(encoded_latent=initial_latent)
        latents = SeedVR2LatentCreator.create_noise_latents(seed=seed, height=initial_latent.shape[-2], width=initial_latent.shape[-1])  # fmt: off

        # 3. Get the pre-computed text embeddings
        txt_pos = SeedVR2TextEmbeddings.load_positive()

        # 4. Create callback context and call before_loop
        ctx = self.callbacks.start(seed=seed, prompt="", config=config)
        ctx.before_loop(latents)

        for t in config.time_steps:
            model_input = mx.concatenate([latents, static_condition], axis=1)

            # 5.t Predict the noise
            noise = self.transformer(
                txt=txt_pos,
                vid=model_input,
                timestep=config.scheduler.timesteps[t],
            )

            # 6.t Take one denoise step
            latents = config.scheduler.step(noise=noise, timestep=t, latents=latents)

            # 7.t Call subscribers in-loop
            ctx.in_loop(t, latents)

            mx.eval(latents)

        # 8. Call subscribers after loop
        ctx.after_loop(latents)

        # 9. Decode the latents and return the image
        decoded = VAEUtil.decode(vae=self.vae, latent=latents, tiling_config=self.tiling_config)
        decoded = decoded[:, :, :true_height, :true_width]
        style = processed_image[:, :, :true_height, :true_width]
        decoded = SeedVR2Util.apply_color_correction(decoded, style)
        return ImageUtil.to_image(
            seed=seed,
            prompt="",
            config=config,
            quantization=self.bits,
            decoded_latents=decoded,
            generation_time=config.time_steps.format_dict["elapsed"],
        )



import mlx.core as mx
from mlx import nn

from mflux.models.seedvr2.model.seedvr2_transformer.swiglu_mlp import SwiGLUMLP


class MMSwiGLU(nn.Module):
    def __init__(
        self,
        vid_dim: int,
        txt_dim: int,
        expand_ratio: int = 4,
        shared_weights: bool = False,
        is_last_layer: bool = False,
    ):
        super().__init__()
        self.shared_weights = shared_weights
        self.is_last_layer = is_last_layer

        if shared_weights:
            self.all = SwiGLUMLP(dim=vid_dim, expand_ratio=expand_ratio)
        else:
            self.vid = SwiGLUMLP(dim=vid_dim, expand_ratio=expand_ratio)
            if not is_last_layer:
                self.txt = SwiGLUMLP(dim=txt_dim, expand_ratio=expand_ratio)

    def __call__(self, vid: mx.array, txt: mx.array) -> tuple[mx.array, mx.array]:
        if self.shared_weights:
            vid_out = self.all(vid)
            txt_out = self.all(txt)
        else:
            vid_out = self.vid(vid)
            txt_out = self.txt(txt) if not self.is_last_layer else txt

        return vid_out, txt_out



import mlx.core as mx
from mlx import nn


class UpSampler(nn.Module):
    def __init__(self, conv_in: int, conv_out: int):
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels=conv_in,
            out_channels=conv_out,
            kernel_size=(3, 3),
            stride=(1, 1),
            padding=(1, 1),
        )

    def __call__(self, input_array: mx.array) -> mx.array:
        input_array = mx.transpose(input_array, (0, 2, 3, 1))
        hidden_states = UpSampler.up_sample_nearest(input_array)
        hidden_state = self.conv(hidden_states)
        return mx.transpose(hidden_state, (0, 3, 1, 2))

    @staticmethod
    def up_sample_nearest(x: mx.array, scale: int = 2):
        B, H, W, C = x.shape
        x = mx.broadcast_to(x[:, :, None, :, None, :], (B, H, scale, W, scale, C))
        x = x.reshape(B, H * scale, W * scale, C)
        return x



from __future__ import annotations

import mlx.core as mx
from mlx import nn
from mlx.core.fast import scaled_dot_product_attention


class QwenAttention(nn.Module):
    def __init__(self, dim: int = 3072, num_heads: int = 24, head_dim: int = 128):
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.head_dim = head_dim
        self.to_q = nn.Linear(dim, dim)
        self.to_k = nn.Linear(dim, dim)
        self.to_v = nn.Linear(dim, dim)
        self.add_q_proj = nn.Linear(dim, dim)
        self.add_k_proj = nn.Linear(dim, dim)
        self.add_v_proj = nn.Linear(dim, dim)
        self.norm_q = nn.RMSNorm(self.head_dim, eps=1e-6)
        self.norm_k = nn.RMSNorm(self.head_dim, eps=1e-6)
        self.norm_added_q = nn.RMSNorm(self.head_dim, eps=1e-6)
        self.norm_added_k = nn.RMSNorm(self.head_dim, eps=1e-6)
        self.attn_to_out = [nn.Linear(dim, dim)]
        self.to_add_out = nn.Linear(dim, dim)

    def __call__(
        self,
        img_modulated: mx.array,
        txt_modulated: mx.array,
        encoder_hidden_states_mask: mx.array | None,
        image_rotary_emb: tuple[mx.array, mx.array],
        block_idx: int | None = None,
    ) -> tuple[mx.array, mx.array]:
        img_query = self.to_q(img_modulated)
        img_key = self.to_k(img_modulated)
        img_value = self.to_v(img_modulated)

        txt_query = self.add_q_proj(txt_modulated)
        txt_key = self.add_k_proj(txt_modulated)
        txt_value = self.add_v_proj(txt_modulated)

        img_query = mx.reshape(img_query, (img_query.shape[0], img_query.shape[1], self.num_heads, self.head_dim))
        img_key = mx.reshape(img_key, (img_key.shape[0], img_key.shape[1], self.num_heads, self.head_dim))
        img_value = mx.reshape(img_value, (img_value.shape[0], img_value.shape[1], self.num_heads, self.head_dim))

        txt_query = mx.reshape(txt_query, (txt_query.shape[0], txt_query.shape[1], self.num_heads, self.head_dim))
        txt_key = mx.reshape(txt_key, (txt_key.shape[0], txt_key.shape[1], self.num_heads, self.head_dim))
        txt_value = mx.reshape(txt_value, (txt_value.shape[0], txt_value.shape[1], self.num_heads, self.head_dim))

        if self.norm_q is not None:
            img_query = self.norm_q(img_query)
        if self.norm_k is not None:
            img_key = self.norm_k(img_key)
        if self.norm_added_q is not None:
            txt_query = self.norm_added_q(txt_query)
        if self.norm_added_k is not None:
            txt_key = self.norm_added_k(txt_key)

        if image_rotary_emb is not None:
            (img_cos, img_sin), (txt_cos, txt_sin) = image_rotary_emb
            img_query = QwenAttention._apply_rope_qwen(img_query, img_cos, img_sin)
            img_key = QwenAttention._apply_rope_qwen(img_key, img_cos, img_sin)
            txt_query = QwenAttention._apply_rope_qwen(txt_query, txt_cos, txt_sin)
            txt_key = QwenAttention._apply_rope_qwen(txt_key, txt_cos, txt_sin)

        joint_query = mx.concatenate([txt_query, img_query], axis=1)
        joint_key = mx.concatenate([txt_key, img_key], axis=1)
        joint_value = mx.concatenate([txt_value, img_value], axis=1)

        seq_txt = txt_modulated.shape[1]
        mask = self._convert_mask_for_qwen(
            mask=encoder_hidden_states_mask,
            joint_seq_len=joint_query.shape[1],
            txt_seq_len=seq_txt,
        )

        hidden_states = self._compute_attention_qwen(
            query=joint_query,
            key=joint_key,
            value=joint_value,
            mask=mask,
            block_idx=block_idx,
        )

        txt_attn_output = hidden_states[:, :seq_txt, :]
        img_attn_output = hidden_states[:, seq_txt:, :]
        img_attn_output = self.attn_to_out[0](img_attn_output)
        txt_attn_output = self.to_add_out(txt_attn_output)

        return img_attn_output, txt_attn_output

    def _compute_attention_qwen(
        self,
        query: mx.array,
        key: mx.array,
        value: mx.array,
        mask: mx.array | None = None,
        block_idx: int | None = None,
    ) -> mx.array:
        query_bhsd = mx.transpose(query, (0, 2, 1, 3))
        key_bhsd = mx.transpose(key, (0, 2, 1, 3))
        value_bhsd = mx.transpose(value, (0, 2, 1, 3))
        head_dim = query.shape[-1]
        scale_value = 1.0 / (head_dim**0.5)
        hidden_states_bhsd = scaled_dot_product_attention(
            query_bhsd, key_bhsd, value_bhsd, scale=scale_value, mask=mask
        )
        hidden_states = mx.transpose(hidden_states_bhsd, (0, 2, 1, 3))
        batch_size = hidden_states.shape[0]
        seq_len = hidden_states.shape[1]
        hidden_states = mx.reshape(hidden_states, (batch_size, seq_len, self.num_heads * self.head_dim))
        hidden_states = hidden_states.astype(query.dtype)
        return hidden_states

    @staticmethod
    def _convert_mask_for_qwen(
        mask: mx.array | None,
        joint_seq_len: int,
        txt_seq_len: int,
    ) -> mx.array | None:
        if mask is None:
            return None

        bsz = mask.shape[0]
        img_seq_len = joint_seq_len - txt_seq_len

        ones_img = mx.ones((bsz, img_seq_len), dtype=mx.float32)
        joint_mask = mx.concatenate([mask.astype(mx.float32), ones_img], axis=1)

        if mx.all(joint_mask >= 0.999):
            return None

        additive = (1.0 - joint_mask) * (-1e9)
        return additive.reshape((additive.shape[0], 1, 1, additive.shape[1]))

    @staticmethod
    def _apply_rope_qwen(x: mx.array, cos_vals: mx.array, sin_vals: mx.array) -> mx.array:
        x_float = x.astype(mx.float32)
        x_reshaped = mx.reshape(x_float, (*x.shape[:-1], -1, 2))

        x_real = x_reshaped[..., 0]
        x_imag = x_reshaped[..., 1]

        freqs_cos = cos_vals[None, :, None, :]
        freqs_sin = sin_vals[None, :, None, :]

        if freqs_cos.shape[-1] != x_real.shape[-1]:
            freqs_cos = freqs_cos[..., : x_real.shape[-1]]
            freqs_sin = freqs_sin[..., : x_real.shape[-1]]

        out_real = x_real * freqs_cos - x_imag * freqs_sin
        out_imag = x_real * freqs_sin + x_imag * freqs_cos

        out_pairs = mx.stack([out_real, out_imag], axis=-1)
        x_out = mx.reshape(out_pairs, (*x.shape[:-1], -1))

        return x_out.astype(x.dtype)



import os
import tempfile
from pathlib import Path
from typing import Callable
from zipfile import ZipFile


class ZipUtil:
    @staticmethod
    def unzip(zip_path: str | Path | None, filename: str, loader: Callable):
        if not zip_path:  # Would be nicer to do this in typing, but that's more effort on the callers' side
            raise ValueError("zip_path cannot be None")
        zip_path = Path(zip_path)
        if not zip_path.exists():
            raise FileNotFoundError(f"ZIP file not found at: {zip_path}")

        if not callable(loader):
            raise ValueError("The file_loader must be a callable (e.g., a function or lambda).")

        with ZipFile(zip_path, "r") as zipf:
            # Normalize files names in the archive by stripping folders
            namelist = {os.path.basename(name): name for name in zipf.namelist()}

            if filename not in namelist:
                raise FileNotFoundError(f"File '{filename}' not found in the ZIP archive.")

            # Use the full path to extract the correct file
            archive_filename = namelist[filename]
            file_data = zipf.read(archive_filename)

            # Create a temporary file to store the extracted content
            temp_file = tempfile.NamedTemporaryFile(suffix=f".{filename.split('.')[-1]}", delete=False)
            try:
                temp_file.write(file_data)
                temp_file.flush()
                temp_file.close()

                # Call the file_loader with the temporary file path
                return loader(temp_file.name)
            finally:
                # Cleanup the temporary file
                if os.path.exists(temp_file.name):
                    os.remove(temp_file.name)

    @staticmethod
    def extract_all(zip_path: str | Path, output_dir: str | Path):
        zip_path = Path(zip_path)
        output_dir = Path(output_dir)

        if not zip_path.exists():
            raise FileNotFoundError(f"ZIP file not found at: {zip_path}")

        if not output_dir.exists():
            output_dir.mkdir(parents=True, exist_ok=True)

        with ZipFile(zip_path, "r") as zipf:
            zipf.extractall(output_dir)

        print(f"All files have been extracted to: {output_dir}")



import mlx.core as mx
from mlx import nn

from mflux.models.qwen.model.qwen_vae.qwen_image_causal_conv_3d import QwenImageCausalConv3D
from mflux.models.qwen.model.qwen_vae.qwen_image_rms_norm import QwenImageRMSNorm


class QwenImageResBlock3D(nn.Module):
    def __init__(self, in_channels: int, out_channels: int):
        super().__init__()
        self.norm1 = QwenImageRMSNorm(in_channels, images=False)
        self.conv1 = QwenImageCausalConv3D(in_channels, out_channels, 3, 1, 1)
        self.norm2 = QwenImageRMSNorm(out_channels, images=False)
        self.conv2 = QwenImageCausalConv3D(out_channels, out_channels, 3, 1, 1)
        if in_channels != out_channels:
            self.skip_conv = QwenImageCausalConv3D(in_channels, out_channels, 1, 1, 0)
        else:
            self.skip_conv = None

    def __call__(self, x: mx.array) -> mx.array:
        residual = x
        x = self.norm1(x)
        x = nn.silu(x)
        x = self.conv1(x)
        x = self.norm2(x)
        x = nn.silu(x)
        x = self.conv2(x)
        if self.skip_conv is not None:
            residual = self.skip_conv(residual)
        return x + residual



import mlx.core as mx
from mlx import nn

from mflux.models.seedvr2.model.seedvr2_transformer.patch_in import PatchIn
from mflux.models.seedvr2.model.seedvr2_transformer.patch_out import PatchOut
from mflux.models.seedvr2.model.seedvr2_transformer.rms_norm import RMSNorm
from mflux.models.seedvr2.model.seedvr2_transformer.time_embedding import TimeEmbedding
from mflux.models.seedvr2.model.seedvr2_transformer.transformer_block import TransformerBlock


class SeedVR2Transformer(nn.Module):
    def __init__(
        self,
        vid_in_channels: int = 33,
        vid_out_channels: int = 16,
        vid_dim: int = 2560,
        txt_in_dim: int = 5120,
        txt_dim: int | None = None,
        emb_dim: int | None = None,
        heads: int = 20,
        head_dim: int = 128,
        expand_ratio: int = 4,
        norm_eps: float = 1e-5,
        patch_size: tuple = (1, 2, 2),
        num_layers: int = 32,
        mm_layers: int = 10,
        rope_dim: int = 128,
        window: tuple[int, int, int] = (4, 3, 3),
    ):
        super().__init__()

        txt_dim = txt_dim if txt_dim is not None else vid_dim
        emb_dim = emb_dim if emb_dim is not None else 6 * vid_dim

        self.vid_dim = vid_dim
        self.txt_dim = txt_dim
        self.emb_dim = emb_dim
        self.num_layers = num_layers
        self.mm_layers = mm_layers

        self.vid_in = PatchIn(
            in_channels=vid_in_channels,
            patch_size=patch_size,
            dim=vid_dim,
        )

        self.txt_in = nn.Linear(txt_in_dim, txt_dim)

        self.emb_in = TimeEmbedding(
            sinusoidal_dim=256,
            hidden_dim=max(vid_dim, txt_dim),
            output_dim=emb_dim,
        )

        self.blocks = []
        for i in range(num_layers):
            shared_weights = i >= mm_layers
            is_last_layer = i == num_layers - 1
            shift = i % 2 == 1

            self.blocks.append(
                TransformerBlock(
                    vid_dim=vid_dim,
                    txt_dim=txt_dim,
                    heads=heads,
                    head_dim=head_dim,
                    expand_ratio=expand_ratio,
                    norm_eps=norm_eps,
                    qk_bias=False,
                    rope_dim=rope_dim,
                    shared_weights=shared_weights,
                    is_last_layer=is_last_layer,
                    window=window,
                    shift=shift,
                )
            )

        self.vid_out_norm = RMSNorm(vid_dim, eps=norm_eps)

        self.out_shift = mx.zeros((vid_dim,))
        self.out_scale = mx.ones((vid_dim,))

        self.vid_out = PatchOut(
            out_channels=vid_out_channels,
            patch_size=patch_size,
            dim=vid_dim,
        )

    def __call__(
        self,
        vid: mx.array,
        txt: mx.array,
        timestep: mx.array,
    ) -> mx.array:
        txt = self.txt_in(txt)
        txt_shape = mx.full((txt.shape[0], 1), txt.shape[1], dtype=mx.int32)
        vid, vid_shape = self.vid_in(vid)
        emb = self.emb_in(timestep)
        emb = emb.reshape(-1, self.vid_dim, 2, 3)

        for block in self.blocks:
            vid, txt = block(
                vid=vid,
                txt=txt,
                emb=emb,
                vid_shape=vid_shape,
                txt_shape=txt_shape,
            )

        vid = self.vid_out_norm(vid)
        vid = self._apply_out_ada(vid, emb=emb)
        vid, vid_shape = self.vid_out(vid, vid_shape)
        return vid

    def _apply_out_ada(self, hidden: mx.array, emb: mx.array) -> mx.array:
        shift_a = emb[:, :, 0, 0][:, None, :]
        scale_a = emb[:, :, 0, 1][:, None, :]
        return hidden * (scale_a + self.out_scale) + (shift_a + self.out_shift)



from typing import List

from mflux.models.common.weights.mapping.weight_mapping import WeightMapping, WeightTarget
from mflux.models.common.weights.mapping.weight_transforms import WeightTransforms


class FIBOVLMWeightMapping(WeightMapping):
    @staticmethod
    def get_vlm_decoder_mapping(num_layers: int = 36) -> List[WeightTarget]:
        return [
            WeightTarget(
                to_pattern="embed_tokens.weight",
                from_pattern=["model.language_model.embed_tokens.weight"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.self_attn.q_proj.weight",
                from_pattern=["model.language_model.layers.{block}.self_attn.q_proj.weight"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.self_attn.q_proj.bias",
                from_pattern=["model.language_model.layers.{block}.self_attn.q_proj.bias"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.self_attn.k_proj.weight",
                from_pattern=["model.language_model.layers.{block}.self_attn.k_proj.weight"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.self_attn.k_proj.bias",
                from_pattern=["model.language_model.layers.{block}.self_attn.k_proj.bias"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.self_attn.v_proj.weight",
                from_pattern=["model.language_model.layers.{block}.self_attn.v_proj.weight"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.self_attn.v_proj.bias",
                from_pattern=["model.language_model.layers.{block}.self_attn.v_proj.bias"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.self_attn.o_proj.weight",
                from_pattern=["model.language_model.layers.{block}.self_attn.o_proj.weight"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.self_attn.o_proj.bias",
                from_pattern=["model.language_model.layers.{block}.self_attn.o_proj.bias"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.self_attn.q_norm.weight",
                from_pattern=["model.language_model.layers.{block}.self_attn.q_norm.weight"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.self_attn.k_norm.weight",
                from_pattern=["model.language_model.layers.{block}.self_attn.k_norm.weight"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.mlp.gate_proj.weight",
                from_pattern=["model.language_model.layers.{block}.mlp.gate_proj.weight"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.mlp.gate_proj.bias",
                from_pattern=["model.language_model.layers.{block}.mlp.gate_proj.bias"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.mlp.up_proj.weight",
                from_pattern=["model.language_model.layers.{block}.mlp.up_proj.weight"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.mlp.up_proj.bias",
                from_pattern=["model.language_model.layers.{block}.mlp.up_proj.bias"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.mlp.down_proj.weight",
                from_pattern=["model.language_model.layers.{block}.mlp.down_proj.weight"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.mlp.down_proj.bias",
                from_pattern=["model.language_model.layers.{block}.mlp.down_proj.bias"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.input_layernorm.weight",
                from_pattern=["model.language_model.layers.{block}.input_layernorm.weight"],
            ),
            WeightTarget(
                to_pattern="layers.{block}.post_attention_layernorm.weight",
                from_pattern=["model.language_model.layers.{block}.post_attention_layernorm.weight"],
            ),
            WeightTarget(
                to_pattern="norm.weight",
                from_pattern=["model.language_model.norm.weight"],
            ),
            WeightTarget(
                to_pattern="lm_head.weight",
                from_pattern=["model.language_model.embed_tokens.weight"],
            ),
        ]

    @staticmethod
    def get_vlm_visual_mapping(depth: int = 24) -> List[WeightTarget]:
        return [
            WeightTarget(
                to_pattern="patch_embed.proj.weight",
                from_pattern=["model.visual.patch_embed.proj.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="patch_embed.proj.bias",
                from_pattern=["model.visual.patch_embed.proj.bias"],
            ),
            WeightTarget(
                to_pattern="pos_embed.weight",
                from_pattern=["model.visual.pos_embed.weight"],
            ),
            WeightTarget(
                to_pattern="blocks.{block}.norm1.weight",
                from_pattern=["model.visual.blocks.{block}.norm1.weight"],
            ),
            WeightTarget(
                to_pattern="blocks.{block}.norm1.bias",
                from_pattern=["model.visual.blocks.{block}.norm1.bias"],
            ),
            WeightTarget(
                to_pattern="blocks.{block}.norm2.weight",
                from_pattern=["model.visual.blocks.{block}.norm2.weight"],
            ),
            WeightTarget(
                to_pattern="blocks.{block}.norm2.bias",
                from_pattern=["model.visual.blocks.{block}.norm2.bias"],
            ),
            # Attention
            WeightTarget(
                to_pattern="blocks.{block}.attn.qkv.weight",
                from_pattern=["model.visual.blocks.{block}.attn.qkv.weight"],
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.qkv.bias",
                from_pattern=["model.visual.blocks.{block}.attn.qkv.bias"],
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.proj.weight",
                from_pattern=["model.visual.blocks.{block}.attn.proj.weight"],
            ),
            WeightTarget(
                to_pattern="blocks.{block}.attn.proj.bias",
                from_pattern=["model.visual.blocks.{block}.attn.proj.bias"],
            ),
            # MLP
            WeightTarget(
                to_pattern="blocks.{block}.mlp.linear_fc1.weight",
                from_pattern=["model.visual.blocks.{block}.mlp.linear_fc1.weight"],
            ),
            WeightTarget(
                to_pattern="blocks.{block}.mlp.linear_fc1.bias",
                from_pattern=["model.visual.blocks.{block}.mlp.linear_fc1.bias"],
            ),
            WeightTarget(
                to_pattern="blocks.{block}.mlp.linear_fc2.weight",
                from_pattern=["model.visual.blocks.{block}.mlp.linear_fc2.weight"],
            ),
            WeightTarget(
                to_pattern="blocks.{block}.mlp.linear_fc2.bias",
                from_pattern=["model.visual.blocks.{block}.mlp.linear_fc2.bias"],
            ),
            # Final merger
            WeightTarget(
                to_pattern="merger.norm.weight",
                from_pattern=["model.visual.merger.norm.weight"],
            ),
            WeightTarget(
                to_pattern="merger.norm.bias",
                from_pattern=["model.visual.merger.norm.bias"],
            ),
            WeightTarget(
                to_pattern="merger.linear_fc1.weight",
                from_pattern=["model.visual.merger.linear_fc1.weight"],
            ),
            WeightTarget(
                to_pattern="merger.linear_fc1.bias",
                from_pattern=["model.visual.merger.linear_fc1.bias"],
            ),
            WeightTarget(
                to_pattern="merger.linear_fc2.weight",
                from_pattern=["model.visual.merger.linear_fc2.weight"],
            ),
            WeightTarget(
                to_pattern="merger.linear_fc2.bias",
                from_pattern=["model.visual.merger.linear_fc2.bias"],
            ),
            WeightTarget(
                to_pattern="deepstack_merger_list.{block}.norm.weight",
                from_pattern=["model.visual.deepstack_merger_list.{block}.norm.weight"],
                max_blocks=3,
            ),
            WeightTarget(
                to_pattern="deepstack_merger_list.{block}.norm.bias",
                from_pattern=["model.visual.deepstack_merger_list.{block}.norm.bias"],
                max_blocks=3,
            ),
            WeightTarget(
                to_pattern="deepstack_merger_list.{block}.linear_fc1.weight",
                from_pattern=["model.visual.deepstack_merger_list.{block}.linear_fc1.weight"],
                max_blocks=3,
            ),
            WeightTarget(
                to_pattern="deepstack_merger_list.{block}.linear_fc1.bias",
                from_pattern=["model.visual.deepstack_merger_list.{block}.linear_fc1.bias"],
                max_blocks=3,
            ),
            WeightTarget(
                to_pattern="deepstack_merger_list.{block}.linear_fc2.weight",
                from_pattern=["model.visual.deepstack_merger_list.{block}.linear_fc2.weight"],
                max_blocks=3,
            ),
            WeightTarget(
                to_pattern="deepstack_merger_list.{block}.linear_fc2.bias",
                from_pattern=["model.visual.deepstack_merger_list.{block}.linear_fc2.bias"],
                max_blocks=3,
            ),
        ]



from mflux.models.common.weights.loading.loaded_weights import LoadedWeights, MetaData
from mflux.models.common.weights.loading.weight_applier import WeightApplier
from mflux.models.common.weights.loading.weight_definition import ComponentDefinition
from mflux.models.common.weights.loading.weight_loader import WeightLoader
from mflux.models.common.weights.saving.model_saver import ModelSaver

__all__ = [
    "ComponentDefinition",
    "LoadedWeights",
    "MetaData",
    "ModelSaver",
    "WeightApplier",
    "WeightLoader",
]



import logging
from typing import TYPE_CHECKING

from mflux.models.common.resolution.actions import ConfigAction, Rule

if TYPE_CHECKING:
    from mflux.models.common.config.model_config import ModelConfig

logger = logging.getLogger(__name__)


class ConfigResolution:
    RULES = frozenset(
        {
            Rule(priority=0, name="exact_match", check="is_exact_match", action=ConfigAction.EXACT_MATCH),
            Rule(priority=1, name="explicit_base", check="has_explicit_base", action=ConfigAction.EXPLICIT_BASE),
            Rule(priority=2, name="infer_substring", check="can_infer_substring", action=ConfigAction.INFER_SUBSTRING),
            Rule(priority=3, name="error", check="always", action=ConfigAction.ERROR),
        }
    )

    @staticmethod
    def resolve(model_name: str, base_model: str | None = None) -> "ModelConfig":
        from mflux.models.common.config.model_config import AVAILABLE_MODELS, ModelConfig
        from mflux.utils.exceptions import InvalidBaseModel, ModelConfigError

        base_models = sorted(
            [m for m in AVAILABLE_MODELS.values() if m.base_model is None],
            key=lambda x: x.priority,
        )

        ctx = {
            "model_name": model_name,
            "base_model": base_model,
            "base_models": base_models,
            "ModelConfig": ModelConfig,
            "InvalidBaseModel": InvalidBaseModel,
            "ModelConfigError": ModelConfigError,
        }

        for rule in sorted(ConfigResolution.RULES, key=lambda r: r.priority):
            if ConfigResolution._check(rule.check, ctx):
                logger.debug(f"Config resolution: '{model_name}' → rule '{rule.name}' ({rule.action.value})")
                return ConfigResolution._execute(rule.action, ctx)

        raise ValueError(f"No rule matched for model_name: {model_name}")

    @staticmethod
    def _check(check: str, ctx: dict) -> bool:
        if check == "is_exact_match":
            model_name = ctx["model_name"]
            for base in ctx["base_models"]:
                if model_name == base.model_name or model_name in base.aliases:
                    return True
            return False
        if check == "has_explicit_base":
            return ctx["base_model"] is not None
        if check == "can_infer_substring":
            model_name_lower = ctx["model_name"].lower()
            for base in ctx["base_models"]:
                for alias in base.aliases:
                    if alias and alias.lower() in model_name_lower:
                        return True
            return False
        if check == "always":
            return True
        return False

    @staticmethod
    def _execute(action: ConfigAction, ctx: dict) -> "ModelConfig":
        model_name = ctx["model_name"]
        base_model = ctx["base_model"]
        base_models = ctx["base_models"]
        InvalidBaseModel = ctx["InvalidBaseModel"]
        ModelConfigError = ctx["ModelConfigError"]

        if action == ConfigAction.EXACT_MATCH:
            for base in base_models:
                if model_name == base.model_name or model_name in base.aliases:
                    return base
            raise ValueError("Exact match check passed but no match found")

        if action == ConfigAction.EXPLICIT_BASE:
            allowed_names = []
            for base in base_models:
                allowed_names.extend(base.aliases + [base.model_name])
            if base_model not in allowed_names:
                raise InvalidBaseModel(f"Invalid base_model. Choose one of {allowed_names}")

            default_base = next(
                (b for b in base_models if base_model == b.model_name or base_model in b.aliases),
                None,
            )
            return ConfigResolution._create_config(model_name, default_base)

        if action == ConfigAction.INFER_SUBSTRING:
            model_name_lower = model_name.lower()
            matching_bases = [
                (b, alias) for b in base_models for alias in b.aliases if alias and alias.lower() in model_name_lower
            ]
            if not matching_bases:
                raise ModelConfigError(f"Cannot infer base_model from {model_name}")

            default_base = sorted(matching_bases, key=lambda x: (-len(x[1]), x[0].priority))[0][0]
            return ConfigResolution._create_config(model_name, default_base)

        if action == ConfigAction.ERROR:
            raise ModelConfigError(f"Cannot infer base_model from {model_name}")

        raise ValueError(f"Unknown action: {action}")

    @staticmethod
    def _create_config(model_name: str, base: "ModelConfig") -> "ModelConfig":
        from mflux.models.common.config.model_config import ModelConfig

        return ModelConfig(
            aliases=base.aliases,
            model_name=model_name,
            base_model=base.model_name,
            controlnet_model=base.controlnet_model,
            custom_transformer_model=base.custom_transformer_model,
            num_train_steps=base.num_train_steps,
            max_sequence_length=base.max_sequence_length,
            supports_guidance=base.supports_guidance,
            requires_sigma_shift=base.requires_sigma_shift,
            priority=base.priority,
        )



import math

import mlx.core as mx
from mlx import nn


class TimeEmbedding(nn.Module):
    def __init__(
        self,
        sinusoidal_dim: int = 256,
        hidden_dim: int = 2560,
        output_dim: int = 15360,
    ):
        super().__init__()
        self.sinusoidal_dim = sinusoidal_dim
        self.proj_in = nn.Linear(sinusoidal_dim, hidden_dim)
        self.proj_hid = nn.Linear(hidden_dim, hidden_dim)
        self.proj_out = nn.Linear(hidden_dim, output_dim)

    def __call__(self, timestep: mx.array) -> mx.array:
        timestep = timestep[None] if timestep.ndim == 0 else timestep
        emb = TimeEmbedding._get_timestep_embedding(
            timesteps=timestep,
            embedding_dim=self.sinusoidal_dim,
        )
        emb = self.proj_in(emb)
        emb = nn.silu(emb)
        emb = self.proj_hid(emb)
        emb = nn.silu(emb)
        emb = self.proj_out(emb)
        return emb

    @staticmethod
    def _get_timestep_embedding(
        timesteps: mx.array,
        embedding_dim: int,
    ) -> mx.array:
        half_dim = embedding_dim // 2
        freqs = mx.exp(mx.arange(half_dim, dtype=mx.float32) * (-math.log(10000) / half_dim))
        args = timesteps[:, None].astype(mx.float32) * freqs
        return mx.concatenate([mx.sin(args), mx.cos(args)], axis=-1)



from typing import Optional, Union

import mlx.core as mx
import numpy as np
from PIL import Image

from mflux.models.common.tokenizer import Tokenizer
from mflux.models.qwen.model.qwen_text_encoder.qwen_prompt_encoder import QwenPromptEncoder
from mflux.models.qwen.model.qwen_text_encoder.qwen_text_encoder import QwenTextEncoder
from mflux.models.qwen.model.qwen_text_encoder.qwen_vision_language_encoder import QwenVisionLanguageEncoder
from mflux.models.qwen.tokenizer.qwen_vision_language_tokenizer import QwenVisionLanguageTokenizer


class QwenVisionLanguagePromptEncoder:
    @staticmethod
    def encode_prompt_with_image(
        prompt: str,
        image: Union[Image.Image, np.ndarray, str],
        negative_prompt: Optional[str] = None,
        prompt_cache: Optional[dict] = None,
        qwen_vl_tokenizer: Optional[QwenVisionLanguageTokenizer] = None,
        qwen_vl_encoder: Optional[QwenVisionLanguageEncoder] = None,
    ) -> tuple[mx.array, mx.array, mx.array, mx.array, mx.array]:
        if qwen_vl_tokenizer is None or qwen_vl_encoder is None:
            raise ValueError("qwen_vl_tokenizer and qwen_vl_encoder are required")

        input_ids, attention_mask, pixel_values, image_grid_thw = qwen_vl_tokenizer.tokenize_with_image(
            prompt=prompt, image=image
        )

        prompt_embeds, prompt_mask = qwen_vl_encoder(
            input_ids=input_ids,
            attention_mask=attention_mask,
            pixel_values=pixel_values,
            image_grid_thw=image_grid_thw,
        )

        if negative_prompt is not None and negative_prompt != "":
            neg_input_ids, neg_attention_mask, neg_pixel_values, neg_image_grid_thw = (
                qwen_vl_tokenizer.tokenize_with_image(prompt=negative_prompt, image=image)
            )

            neg_prompt_embeds, neg_prompt_mask = qwen_vl_encoder(
                input_ids=neg_input_ids,
                attention_mask=neg_attention_mask,
                pixel_values=neg_pixel_values,
                image_grid_thw=neg_image_grid_thw,
            )
        else:
            # If no negative prompt provided, mirror shapes with zeros masks/embeds to disable CFG cleanly
            neg_prompt_embeds, neg_prompt_mask = prompt_embeds, prompt_mask

        # 3. Return the result including image_grid_thw for conditioning
        result = (prompt_embeds, prompt_mask, neg_prompt_embeds, neg_prompt_mask, image_grid_thw)
        return result

    @staticmethod
    def encode_prompt_auto(
        prompt: str,
        negative_prompt: Optional[str] = None,
        image: Optional[Union[Image.Image, np.ndarray, str]] = None,
        prompt_cache: Optional[dict] = None,
        qwen_tokenizer: Optional[Tokenizer] = None,
        qwen_text_encoder: Optional[QwenTextEncoder] = None,
        qwen_vl_tokenizer: Optional[QwenVisionLanguageTokenizer] = None,
        qwen_vl_encoder: Optional[QwenVisionLanguageEncoder] = None,
    ) -> tuple[mx.array, mx.array, mx.array, mx.array]:
        if prompt_cache is None:
            prompt_cache = {}

        if image is not None and qwen_vl_tokenizer is not None and qwen_vl_encoder is not None:
            # Use vision-language encoding for edit tasks
            return QwenVisionLanguagePromptEncoder.encode_prompt_with_image(
                prompt=prompt,
                image=image,
                negative_prompt=negative_prompt,
                prompt_cache=prompt_cache,
                qwen_vl_tokenizer=qwen_vl_tokenizer,
                qwen_vl_encoder=qwen_vl_encoder,
            )
        else:
            # Fall back to text-only encoding for regular generation
            if qwen_tokenizer is None or qwen_text_encoder is None:
                raise ValueError(
                    "Text-only components (qwen_tokenizer, qwen_text_encoder) are required when no image is provided"
                )

            return QwenPromptEncoder.encode_prompt(
                prompt=prompt,
                negative_prompt=negative_prompt or "",
                prompt_cache=prompt_cache,
                qwen_tokenizer=qwen_tokenizer,
                qwen_text_encoder=qwen_text_encoder,
            )



from typing import TYPE_CHECKING

import mlx.core as mx

if TYPE_CHECKING:
    from mflux.models.common.config.config import Config

from mflux.models.common.schedulers.base_scheduler import BaseScheduler


class SeedVR2EulerScheduler(BaseScheduler):
    def __init__(self, config: "Config"):
        self.config = config
        self.num_inference_steps = config.num_inference_steps
        self.num_train_timesteps = config.num_train_steps if config.num_train_steps is not None else 1000
        self.cfg_scale = config.guidance
        self.T = float(self.num_train_timesteps)
        self._timesteps, self._sigmas = self._compute_timesteps_and_sigmas()

    @property
    def timesteps(self) -> mx.array:
        return self._timesteps

    @property
    def sigmas(self) -> mx.array:
        return self._sigmas

    def _compute_timesteps_and_sigmas(self) -> tuple[mx.array, mx.array]:
        step_size = self.T / self.num_inference_steps
        timesteps = []
        for i in range(self.num_inference_steps + 1):
            t = self.T - i * step_size
            timesteps.append(max(t, 0.0))
        timesteps_arr = mx.array(timesteps, dtype=mx.float32)
        sigmas_arr = timesteps_arr / self.T
        return timesteps_arr, sigmas_arr

    def step(
        self,
        noise: mx.array,
        timestep: int,
        latents: mx.array,
        **kwargs,
    ) -> mx.array:
        model_output = noise
        sample = latents
        timestep_idx = timestep
        t = self._timesteps[timestep_idx]
        s = self._timesteps[timestep_idx + 1]
        t_norm = t / self.T
        s_norm = s / self.T
        pred_x_0 = sample - t_norm * model_output
        pred_noise = sample + (1 - t_norm) * model_output
        if s > 0:
            next_sample = (1 - s_norm) * pred_x_0 + s_norm * pred_noise
        else:
            next_sample = pred_x_0
        return next_sample
